import json
import pytest

from indy.anoncreds import issuer_create_and_store_credential_def
from indy.ledger import build_get_cred_def_request, build_cred_def_request

from plenum.common.exceptions import RequestNackedException
from plenum.common.constants import TXN_METADATA, TXN_METADATA_ID
from plenum.test.helper import sdk_sign_and_submit_req, sdk_get_and_check_replies

from indy_node.test.helper import modify_field


@pytest.fixture(scope="module")
def added_claim_def_id(looper, sdk_pool_handle, nodeSet,
                       sdk_wallet_trustee, schema_json):
    wallet_handle, identifier = sdk_wallet_trustee
    _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
        wallet_handle, identifier, schema_json, "some_tag", "CL", json.dumps({"support_revocation": True})))
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    rep = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    return rep[0][1]['result'][TXN_METADATA][TXN_METADATA_ID]


@pytest.fixture(scope="module")
def added_claim_def_id_large_schema(looper, sdk_pool_handle, nodeSet,
                                    sdk_wallet_trustee, large_schema_json):
    wallet_handle, identifier = sdk_wallet_trustee
    _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
        wallet_handle, identifier, large_schema_json, "some_tag", "CL", json.dumps({"support_revocation": True})))
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    rep = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    return rep[0][1]['result'][TXN_METADATA][TXN_METADATA_ID]


def test_send_get_claim_def_succeeds(looper, sdk_pool_handle, nodeSet,
                                     sdk_wallet_trustee, added_claim_def_id):
    _, did = sdk_wallet_trustee
    request = looper.loop.run_until_complete(build_get_cred_def_request(did, added_claim_def_id))
    sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])


def test_send_get_claim_def_succeeds_for_large_schema(looper, sdk_pool_handle, nodeSet,
                                                      sdk_wallet_trustee, added_claim_def_id_large_schema):
    _, did = sdk_wallet_trustee
    request = looper.loop.run_until_complete(build_get_cred_def_request(did, added_claim_def_id_large_schema))
    sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])


def test_send_get_claim_def_as_client_succeeds(
        looper,
        sdk_pool_handle,
        nodeSet,
        added_claim_def_id,
        sdk_wallet_client):
    _, did = sdk_wallet_client
    request = looper.loop.run_until_complete(build_get_cred_def_request(did, added_claim_def_id))
    sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_client, request)])


def test_send_get_claim_def_with_invalid_ref_fails(looper, sdk_pool_handle, nodeSet,
                                                   sdk_wallet_trustee, added_claim_def_id):
    _, did = sdk_wallet_trustee
    request = looper.loop.run_until_complete(build_get_cred_def_request(did, added_claim_def_id))
    request = modify_field(request, '!@#', 'operation', 'ref')
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    e.match('expected types \'int\', got \'str\'')


def test_send_get_claim_def_with_invalid_signature_not_get_claim(
        looper, sdk_pool_handle, nodeSet, sdk_wallet_trustee, added_claim_def_id):
    _, did = sdk_wallet_trustee
    request = looper.loop.run_until_complete(build_get_cred_def_request(did, added_claim_def_id))
    request = modify_field(request, 'ABC', 'operation', 'signature_type')
    rep = sdk_get_and_check_replies(looper, [sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request)])
    assert rep[0][1]['result']['data'] is None



import pytest

from indy_common.authorize.auth_actions import AuthActionEdit
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_common.types import Request

action = AuthActionEdit(txn_type="some_type",
                        field='some_field',
                        old_value='old_value',
                        new_value='new_value')


@pytest.fixture(scope='module')
def write_auth_req_validator(write_auth_req_validator):
    write_auth_req_validator.auth_map[action.get_action_id()] = AuthConstraint(role="*",
                                                                               sig_count=0)
    return write_auth_req_validator


def test_auth_request_without_signatures(write_request_validation):
    req = Request(identifier="UnknownIdr",
                  operation={})
    assert write_request_validation(req,
                                    [action])



from plenum.test.helper import sdk_sign_and_submit_op, sdk_get_and_check_replies


def sdk_suspend_role(looper, sdk_pool_handle, sdk_wallet_sender, susp_did):
    op = {'type': '1',
          'dest': susp_did,
          'role': None}
    req = sdk_sign_and_submit_op(looper, sdk_pool_handle, sdk_wallet_sender, op)
    sdk_get_and_check_replies(looper, [req])



from plenum.common.transactions import Transactions, PlenumTransactions


class IndyTransactions(Transactions):
    #  These numeric constants CANNOT be changed once they have been used,
    #  because that would break backwards compatibility with the ledger
    #  Also the numeric constants CANNOT collide with transactions in plenum
    NODE = PlenumTransactions.NODE.value
    NYM = PlenumTransactions.NYM.value
    ATTRIB = "100"
    SCHEMA = "101"
    CLAIM_DEF = "102"

    DISCLO = "103"
    GET_ATTR = "104"
    GET_NYM = "105"
    GET_TXNS = PlenumTransactions.GET_TXN.value
    GET_SCHEMA = "107"
    GET_CLAIM_DEF = "108"

    POOL_UPGRADE = "109"
    NODE_UPGRADE = "110"

    POOL_CONFIG = "111"

    CHANGE_KEY = "112"

    REVOC_REG_DEF = "113"
    REVOC_REG_ENTRY = "114"
    GET_REVOC_REG_DEF = "115"
    GET_REVOC_REG = "116"
    GET_REVOC_REG_DELTA = "117"

    POOL_RESTART = "118"
    VALIDATOR_INFO = "119"

    AUTH_RULE = "120"
    GET_AUTH_RULE = "121"
    AUTH_RULES = "122"

    # Rich Schema
    JSON_LD_CONTEXT = "200"
    RICH_SCHEMA = "201"
    RICH_SCHEMA_ENCODING = "202"
    RICH_SCHEMA_MAPPING = "203"
    RICH_SCHEMA_CRED_DEF = "204"
    RICH_SCHEMA_PRES_DEF = "205"
    GET_RICH_SCHEMA_OBJECT_BY_ID = "300"
    GET_RICH_SCHEMA_OBJECT_BY_METADATA = "301"

    @staticmethod
    def get_name_from_code(code: str):
        try:
            return IndyTransactions(code).name
        except ValueError:
            return "Unknown_transaction_type"



import pytest

from indy_node.test.upgrade.helper import sdk_change_bls_key
from plenum.test.bls.helper import check_bls_key


@pytest.fixture(scope="module")
def update_bls_keys(looper, tconf, nodeSet, sdk_pool_handle,
                    sdk_wallet_stewards):
    node = nodeSet[0]
    new_blspk = sdk_change_bls_key(looper, nodeSet, node,
                                   sdk_pool_handle, sdk_wallet_stewards[0])

    check_bls_key(new_blspk, node, nodeSet)


def test_node_schedules_upgrade_after_bls_keys_update(update_bls_keys,
                                                      upgradeScheduled):
    # Upgrade should work even after an update to the pool ledger with a
    # transaction that does not contain `SERVICES` field
    pass



from collections import OrderedDict
from indy_common.types import ClientGetRevocRegDeltaField
from plenum.common.messages.fields import NonEmptyStringField, ConstantField, IntegerField
from indy_common.constants import TXN_TYPE, REVOC_REG_DEF_ID, FROM, TO, GET_REVOC_REG_DELTA


EXPECTED_GET_REVOC_REG_DELTA_FIELD = OrderedDict([
    (TXN_TYPE, ConstantField(GET_REVOC_REG_DELTA)),
    (REVOC_REG_DEF_ID, NonEmptyStringField),
    (FROM, IntegerField),
    (TO, IntegerField),
])


def test_get_revoc_reg_def_schema():
    actual_field_name = OrderedDict(ClientGetRevocRegDeltaField.schema).keys()
    assert EXPECTED_GET_REVOC_REG_DELTA_FIELD.keys() == actual_field_name



from collections import OrderedDict

import pytest

from indy_common.types import ClientJsonLdContextOperation, ClientRichSchemaOperation, \
    ClientRichSchemaEncodingOperation, ClientRichSchemaMappingOperation, ClientRichSchemaCredDefOperation, \
    ClientRichSchemaPresDefOperation
from plenum.common.messages.fields import ConstantField, LimitedLengthStringField, NonEmptyStringField, VersionField

EXPECTED_ORDERED_FIELDS = OrderedDict([
    ("type", ConstantField),
    ("ver", LimitedLengthStringField),
    ("id", NonEmptyStringField),
    ("rsType", ConstantField),
    ("rsName", LimitedLengthStringField),
    ("rsVersion", VersionField),
    ("content", NonEmptyStringField),
])


@pytest.mark.parametrize('operation_class',
                         [ClientJsonLdContextOperation, ClientRichSchemaOperation, ClientRichSchemaEncodingOperation,
                          ClientRichSchemaMappingOperation,
                          ClientRichSchemaCredDefOperation,
                          ClientRichSchemaPresDefOperation])
def test_has_expected_fields(operation_class):
    actual_field_names = OrderedDict(operation_class.schema).keys()
    assert actual_field_names == EXPECTED_ORDERED_FIELDS.keys()


@pytest.mark.parametrize('operation_class, txn_type, rs_type',
                         [(ClientJsonLdContextOperation, "200", 'ctx'),
                          (ClientRichSchemaOperation, "201", 'sch'),
                          (ClientRichSchemaEncodingOperation, "202", 'enc'),
                          (ClientRichSchemaMappingOperation, "203", 'map'),
                          (ClientRichSchemaCredDefOperation, "204", 'cdf'),
                          (ClientRichSchemaPresDefOperation, "205", 'pdf')])
def test_has_expected_validators(operation_class, txn_type, rs_type):
    schema = dict(operation_class.schema)
    for field, validator in EXPECTED_ORDERED_FIELDS.items():
        assert isinstance(schema[field], validator)
    assert schema["rsType"].value == rs_type
    assert schema["type"].value == txn_type



import json

import pytest
from indy.did import replace_keys_start

from indy_common.authorize.auth_actions import EDIT_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_common.constants import NYM
from indy_node.test.auth_rule.auth_framework.basic import AuthTest
from indy_node.test.helper import sdk_rotate_verkey
from plenum.common.constants import TRUSTEE, VERKEY
from plenum.common.exceptions import RequestRejectedException
from plenum.test.pool_transactions.helper import sdk_add_new_nym

from indy_node.test.helper import build_auth_rule_request_json


class RotateKeyTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)
        self.creator_wallet = env.sdk_wallet_trustee
        self.test_nym = None

    def prepare(self):
        self.test_nym = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.creator_wallet, role=None)
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()

    def run(self):
        wh, client_did = self.test_nym
        _, trustee_did = self.creator_wallet

        # Step 1. Check default auth rule
        sdk_rotate_verkey(self.looper, self.sdk_pool_handle, wh, client_did, client_did)
        verkey = self.sdk_modified_verkey_rotate_failed(self.sdk_pool_handle, wh, trustee_did, client_did)

        # Step 2. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 3. Check, that we cannot add new steward by old way
        sdk_rotate_verkey(self.looper, self.sdk_pool_handle, wh, trustee_did, client_did, verkey)
        verkey = self.sdk_modified_verkey_rotate_failed(self.sdk_pool_handle, wh, client_did, client_did)

        # Step 4. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        sdk_rotate_verkey(self.looper, self.sdk_pool_handle, wh, client_did, client_did, verkey)
        self.sdk_modified_verkey_rotate_failed(self.sdk_pool_handle, wh, trustee_did, client_did)

    def result(self):
        pass

    def get_changed_auth_rule(self):
        constraint = AuthConstraint(role=TRUSTEE,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.creator_wallet[1],
            auth_action=EDIT_PREFIX,
            auth_type=NYM,
            field=VERKEY,
            old_value='*',
            new_value='*',
            constraint=constraint.as_dict
        )

    def sdk_modified_verkey_rotate_failed(self, sdk_pool_handle, wh,
                                          did_of_changer,
                                          did_of_changed):
        verkey = self.looper.loop.run_until_complete(
            replace_keys_start(wh, did_of_changed, json.dumps({})))

        with pytest.raises(RequestRejectedException):
            sdk_add_new_nym(self.looper, sdk_pool_handle,
                            (wh, did_of_changer), dest=did_of_changed,
                            verkey=verkey)
        return verkey



import json

import pytest

from indy_common.constants import JSON_LD_CONTEXT, RICH_SCHEMA, RICH_SCHEMA_ENCODING, \
    RICH_SCHEMA_MAPPING, RICH_SCHEMA_CRED_DEF, \
    RS_CONTEXT_TYPE_VALUE, RS_SCHEMA_TYPE_VALUE, RS_ENCODING_TYPE_VALUE, RS_MAPPING_TYPE_VALUE, RS_CRED_DEF_TYPE_VALUE, \
    GET_RICH_SCHEMA_OBJECT_BY_ID, RS_ID, GET_RICH_SCHEMA_OBJECT_BY_METADATA, RS_NAME, RS_VERSION, RS_TYPE, \
    RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE
from indy_node.test.api.helper import sdk_write_rich_schema_object_and_check
from indy_node.test.helper import rich_schemas_enabled_scope
from indy_node.test.rich_schema.templates import RICH_SCHEMA_EX1, W3C_BASE_CONTEXT, RICH_SCHEMA_PRES_DEF_EX1, \
    RICH_SCHEMA_CRED_DEF_EX1, RICH_SCHEMA_MAPPING_EX1, RICH_SCHEMA_ENCODING_EX1
from indy_node.test.state_proof.helper import check_valid_proof, sdk_submit_operation_and_get_result
from plenum.common.constants import TXN_TYPE
from plenum.common.exceptions import RequestNackedException
from plenum.common.util import randomString, SortedDict


@pytest.fixture(scope="module")
def tconf(tconf):
    with rich_schemas_enabled_scope(tconf):
        yield tconf

# The order of creation is essential as some rich schema object reference others by ID
# Encoding's id must be equal to the one used in RICH_SCHEMA_MAPPING_EX1

# txn_type, rs_type, content, rs_id, rs_name, rs_version:
PARAMS = [
    (
        JSON_LD_CONTEXT, RS_CONTEXT_TYPE_VALUE, W3C_BASE_CONTEXT, randomString(), randomString(), '1.0'
    ),
    (
        RICH_SCHEMA, RS_SCHEMA_TYPE_VALUE, RICH_SCHEMA_EX1, RICH_SCHEMA_EX1['@id'], randomString(), '2.0'
    ),
    (
        RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE, RICH_SCHEMA_ENCODING_EX1,
        "did:sov:1x9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD", randomString(), '1.1'
    ),
    (
        RICH_SCHEMA_MAPPING, RS_MAPPING_TYPE_VALUE, RICH_SCHEMA_MAPPING_EX1, RICH_SCHEMA_MAPPING_EX1['@id'],
        randomString(),
        '10.0'
    ),
    (
        RICH_SCHEMA_CRED_DEF, RS_CRED_DEF_TYPE_VALUE, RICH_SCHEMA_CRED_DEF_EX1, randomString(),
        randomString(), '1000.0'
    ),
    (
        RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE, RICH_SCHEMA_PRES_DEF_EX1, RICH_SCHEMA_PRES_DEF_EX1['@id'],
        randomString(), '1.1000.1'
    )
]


@pytest.fixture(scope='module')
def write_rich_schema(looper, sdk_pool_handle, sdk_wallet_endorser):
    for txn_type, rs_type, content, rs_id, rs_name, rs_version in PARAMS:
        sdk_write_rich_schema_object_and_check(looper, sdk_wallet_endorser, sdk_pool_handle,
                                               txn_type=txn_type, rs_id=rs_id, rs_name=rs_name,
                                               rs_version=rs_version, rs_type=rs_type, rs_content=content)


@pytest.mark.parametrize('txn_type, rs_type, content, rs_id, rs_name, rs_version', PARAMS)
def test_send_get_rich_schema_obj_by_id(looper, sdk_pool_handle, sdk_wallet_endorser,
                                        write_rich_schema,
                                        txn_type, rs_type, content, rs_id, rs_name, rs_version):
    get_rich_schema_by_id_operation = {
        TXN_TYPE: GET_RICH_SCHEMA_OBJECT_BY_ID,
        RS_ID: rs_id,
    }

    result = sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                                 sdk_wallet_endorser,
                                                 get_rich_schema_by_id_operation)
    expected_data = SortedDict({
        'id': rs_id,
        'rsType': rs_type,
        'rsName': rs_name,
        'rsVersion': rs_version,
        'content': json.dumps(content),
        'from': sdk_wallet_endorser[1],
        'endorser': None,
        'ver': None
    })
    assert SortedDict(result['data']) == expected_data
    assert result['seqNo']
    assert result['txnTime']
    assert result['state_proof']
    check_valid_proof(result)


@pytest.mark.parametrize('txn_type, rs_type, content, rs_id, rs_name, rs_version', PARAMS)
def test_send_get_rich_schema_obj_by_metadata(looper, sdk_pool_handle, sdk_wallet_endorser, sdk_wallet_client,
                                              write_rich_schema,
                                              txn_type, rs_type, content, rs_id, rs_name, rs_version):
    get_rich_schema_by_metadata_operation = {
        TXN_TYPE: GET_RICH_SCHEMA_OBJECT_BY_METADATA,
        RS_NAME: rs_name,
        RS_VERSION: rs_version,
        RS_TYPE: rs_type
    }

    result = sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                                 sdk_wallet_client,
                                                 get_rich_schema_by_metadata_operation)
    expected_data = SortedDict({
        'id': rs_id,
        'rsType': rs_type,
        'rsName': rs_name,
        'rsVersion': rs_version,
        'content': json.dumps(content),
        'from': sdk_wallet_endorser[1],
        'endorser': None,
        'ver': None
    })
    assert SortedDict(result['data']) == expected_data
    assert result['seqNo']
    assert result['txnTime']
    assert result['state_proof']
    check_valid_proof(result)


@pytest.mark.parametrize('txn_type, rs_type, content, rs_id, rs_name, rs_version', PARAMS)
def test_send_get_rich_schema_obj_by_invalid_id(looper, sdk_pool_handle, sdk_wallet_endorser,
                                                write_rich_schema,
                                                txn_type, rs_type, content, rs_id, rs_name, rs_version):
    get_rich_schema_by_id_operation = {
        TXN_TYPE: GET_RICH_SCHEMA_OBJECT_BY_ID,
        RS_ID: randomString(),
    }

    result = sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                                 sdk_wallet_endorser,
                                                 get_rich_schema_by_id_operation)
    assert result['data'] is None
    assert result['seqNo'] is None
    assert result['txnTime'] is None
    assert result['state_proof']
    check_valid_proof(result)


@pytest.mark.parametrize('txn_type, rs_type, content, rs_id, rs_name, rs_version', PARAMS)
@pytest.mark.parametrize('invalid_meta_name, invalid_meta_value', [(RS_NAME, randomString()),
                                                                   (RS_VERSION, "100.3"),
                                                                   (RS_TYPE, randomString())])
def test_send_get_rich_schema_obj_by_invalid_metadata(looper, sdk_pool_handle, sdk_wallet_endorser,
                                                      write_rich_schema,
                                                      txn_type, rs_type, content, rs_id, rs_name, rs_version,
                                                      invalid_meta_name, invalid_meta_value):
    get_rich_schema_by_metadata_operation = {
        TXN_TYPE: GET_RICH_SCHEMA_OBJECT_BY_METADATA,
        RS_NAME: rs_name,
        RS_VERSION: rs_version,
        RS_TYPE: rs_type
    }
    get_rich_schema_by_metadata_operation[invalid_meta_name] = invalid_meta_value

    result = sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                                 sdk_wallet_endorser,
                                                 get_rich_schema_by_metadata_operation)
    assert result['data'] is None
    assert result['seqNo'] is None
    assert result['txnTime'] is None
    assert result['state_proof']
    check_valid_proof(result)


@pytest.mark.parametrize('txn_type, rs_type, content, rs_id, rs_name, rs_version', PARAMS)
def test_send_get_rich_schema_obj_by_no_id(looper, sdk_pool_handle, sdk_wallet_endorser,
                                           write_rich_schema,
                                           txn_type, rs_type, content, rs_id, rs_name, rs_version):
    get_rich_schema_by_id_operation = {
        TXN_TYPE: GET_RICH_SCHEMA_OBJECT_BY_ID,
    }

    with pytest.raises(RequestNackedException, match='missed fields'):
        sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                            sdk_wallet_endorser,
                                            get_rich_schema_by_id_operation)


@pytest.mark.parametrize('txn_type, rs_type, content, rs_id, rs_name, rs_version', PARAMS)
@pytest.mark.parametrize('absent_meta', [RS_NAME, RS_VERSION, RS_TYPE])
def test_send_get_rich_schema_obj_by_no_metadata(looper, sdk_pool_handle, sdk_wallet_endorser,
                                                 write_rich_schema,
                                                 txn_type, rs_type, content, rs_id, rs_name, rs_version,
                                                 absent_meta):
    get_rich_schema_by_metadata_operation = {
        TXN_TYPE: GET_RICH_SCHEMA_OBJECT_BY_METADATA,
        RS_NAME: rs_name,
        RS_VERSION: rs_version,
        RS_TYPE: rs_type
    }
    get_rich_schema_by_metadata_operation.pop(absent_meta, None)

    with pytest.raises(RequestNackedException, match='missed fields'):
        sdk_submit_operation_and_get_result(looper, sdk_pool_handle,
                                            sdk_wallet_endorser,
                                            get_rich_schema_by_metadata_operation)



from hashlib import sha256
import json
from random import randint
import time
from typing import Optional

from indy.ledger import build_get_attrib_request
from libnacl.secret import SecretBox
from plenum.common.exceptions import RequestNackedException
from plenum.test.helper import sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request
import pytest

from indy_node.test.helper import (
    createUuidIdentifier,
    modify_field,
    sdk_add_attribute_and_check,
    sdk_get_attribute_and_check,
)

attrib_name = 'dateOfBirth'

secretBox = SecretBox()
enc_data = secretBox.encrypt(json.dumps({'name': 'Alice'}).encode()).hex()
hash_data = sha256(json.dumps({'name': 'Alice'}).encode()).hexdigest()


@pytest.fixture(scope="module")
def send_raw_attrib(looper, sdk_pool_handle, sdk_wallet_trustee):
    rep = sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_trustee,
                                      json.dumps({
                                          attrib_name: {
                                              'dayOfMonth': 23,
                                              'year': 1984,
                                              'month': 5
                                          }
                                      }))

    return rep


@pytest.fixture
def send_raw_attrib_factory(looper, sdk_pool_handle, sdk_wallet_trustee):

    def _factory(attrib: dict):
        rep = sdk_add_attribute_and_check(
            looper, sdk_pool_handle, sdk_wallet_trustee,
            json.dumps(attrib)
        )

        return rep
    return _factory


@pytest.fixture(scope="module")
def send_enc_attrib(looper, sdk_pool_handle, sdk_wallet_trustee):
    rep = sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_trustee, None,
                                      enc=json.dumps({attrib_name: enc_data}))
    return rep


@pytest.fixture(scope="module")
def send_hash_attrib(looper, sdk_pool_handle, sdk_wallet_trustee):
    rep = sdk_add_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_trustee, None,
                                      xhash=hash_data)
    return rep


def test_send_get_attr_succeeds_for_existing_uuid_dest(
        looper, sdk_pool_handle, sdk_wallet_trustee, send_raw_attrib):
    wh, did = sdk_wallet_trustee
    sdk_get_attribute_and_check(looper, sdk_pool_handle, sdk_wallet_trustee, did, attrib_name)


def test_send_get_attr_fails_for_nonexistent_uuid_dest(
        looper, sdk_pool_handle, sdk_wallet_trustee, send_raw_attrib):
    _, submitter_did = sdk_wallet_trustee
    req = looper.loop.run_until_complete(
        build_get_attrib_request(submitter_did, submitter_did, attrib_name, None, None))
    req = modify_field(req, submitter_did[:-10], 'operation', 'dest')
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, req)
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [request_couple])
    e.match('should be one of \[16, 32\]')


def test_send_get_attr_fails_for_invalid_attrib(
        looper, sdk_pool_handle, sdk_wallet_trustee, send_raw_attrib):
    did = createUuidIdentifier()
    _, submitter_did = sdk_wallet_trustee
    req = looper.loop.run_until_complete(
        build_get_attrib_request(submitter_did, did, attrib_name, None, None))
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, req)
    sdk_get_and_check_replies(looper, [request_couple])


def test_send_get_attr_fails_with_missing_dest(
        looper, sdk_pool_handle, sdk_wallet_trustee, send_raw_attrib):
    _, submitter_did = sdk_wallet_trustee
    req = looper.loop.run_until_complete(
        build_get_attrib_request(submitter_did, submitter_did, attrib_name, None, None))
    req = modify_field(req, '', 'operation', 'dest')
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, req)
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [request_couple])
    e.match('should be one of \[16, 32\]')


def test_send_get_attr_fails_with_missing_attrib(
        looper, sdk_pool_handle, sdk_wallet_trustee, send_raw_attrib):
    _, submitter_did = sdk_wallet_trustee
    req = looper.loop.run_until_complete(
        build_get_attrib_request(submitter_did, submitter_did, attrib_name, None, None))
    req = json.loads(req)
    del req['operation']['raw']
    req = json.dumps(req)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, req)
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [request_couple])
    e.match('missed fields')


def test_send_get_attr_enc_succeeds_for_existing_uuid_dest(
        looper, sdk_pool_handle, sdk_wallet_trustee, send_enc_attrib):
    _, submitter_did = sdk_wallet_trustee
    req = looper.loop.run_until_complete(
        build_get_attrib_request(submitter_did, submitter_did, None, None, attrib_name))
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, req)
    sdk_get_and_check_replies(looper, [request_couple])


def test_send_get_attr_hash_succeeds_for_existing_uuid_dest(
        looper, sdk_pool_handle, sdk_wallet_trustee, send_hash_attrib):
    _, submitter_did = sdk_wallet_trustee
    req = looper.loop.run_until_complete(
        build_get_attrib_request(submitter_did, submitter_did, None, hash_data, None))
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, req)
    sdk_get_and_check_replies(looper, [request_couple])


def test_get_attr_by_timestamp(
    looper, sdk_pool_handle, sdk_wallet_trustee, send_raw_attrib_factory
):
    _, did = sdk_wallet_trustee

    # Setup
    initial = send_raw_attrib_factory({"attrib": 1})
    time.sleep(3)
    final = send_raw_attrib_factory({"attrib": 2})

    timestamp = initial[0][1]["result"]["txnMetadata"]["txnTime"]
    update_timestamp = final[0][1]["result"]["txnMetadata"]["txnTime"]

    def _get_attrib(timestamp: Optional[int] = None):
        raw_req = looper.loop.run_until_complete(
            build_get_attrib_request(did, did, "attrib", None, None))

        req = json.loads(raw_req)
        if timestamp:
            req["operation"]["timestamp"] = timestamp

        request_couple = sdk_sign_and_send_prepared_request(
            looper, sdk_wallet_trustee,
            sdk_pool_handle, json.dumps(req)
        )
        replies = sdk_get_and_check_replies(looper, [request_couple])
        return json.loads(replies[0][1]["result"]["data"])["attrib"]

    assert _get_attrib() == 2
    assert _get_attrib(timestamp=timestamp) == 1
    assert _get_attrib(randint(timestamp + 1, update_timestamp - 1)) == 1


def test_get_attr_by_seq_no(
    looper, sdk_pool_handle, sdk_wallet_trustee, send_raw_attrib_factory
):
    _, did = sdk_wallet_trustee

    # Setup
    initial = send_raw_attrib_factory({"attrib": 1})
    time.sleep(3)
    send_raw_attrib_factory({"attrib": 2})

    seq_no = initial[0][1]["result"]["txnMetadata"]["seqNo"]

    def _get_attrib(seq_no: Optional[int] = None):
        raw_req = looper.loop.run_until_complete(
            build_get_attrib_request(did, did, "attrib", None, None))

        req = json.loads(raw_req)
        if seq_no:
            req["operation"]["seqNo"] = seq_no

        request_couple = sdk_sign_and_send_prepared_request(
            looper, sdk_wallet_trustee,
            sdk_pool_handle, json.dumps(req)
        )
        replies = sdk_get_and_check_replies(looper, [request_couple])
        return json.loads(replies[0][1]["result"]["data"])["attrib"]

    assert _get_attrib() == 2
    assert _get_attrib(seq_no=seq_no) == 1



from indy_common.config_util import getConfig
from plenum.common.constants import TRUSTEE, STEWARD, NODE, LEDGERS_FREEZE
from stp_core.common.log import getlogger

from indy_common.constants import OWNER, POOL_UPGRADE, ENDORSER, NYM, \
    POOL_CONFIG, SCHEMA, CLAIM_DEF, \
    POOL_RESTART, VALIDATOR_INFO, NETWORK_MONITOR
from indy_common.roles import Roles

logger = getlogger()


# TODO: make this class the only point of authorization and checking permissions!
# There are some duplicates of this logic in *_req_handler classes

def generate_auth_map(valid_roles):
    auth_map = {
        '{}_role__{}'.format(NYM, TRUSTEE):
            {TRUSTEE: []},
        '{}_role__{}'.format(NYM, STEWARD):
            {TRUSTEE: []},
        '{}_role__{}'.format(NYM, ENDORSER):
            {TRUSTEE: [], STEWARD: []},
        '{}_role__'.format(NYM):
            {TRUSTEE: [], STEWARD: [], ENDORSER: []},
        '{}_role_{}_'.format(NYM, TRUSTEE):
            {TRUSTEE: []},
        '{}_role_{}_'.format(NYM, STEWARD):
            {TRUSTEE: []},
        '{}_role_{}_'.format(NYM, ENDORSER):
            {TRUSTEE: []},
        '{}_<any>_<any>_<any>'.format(SCHEMA):
            {TRUSTEE: [], STEWARD: [], ENDORSER: []},
        '{}_<any>_<any>_<any>'.format(CLAIM_DEF):
            {TRUSTEE: [OWNER, ], STEWARD: [OWNER, ], ENDORSER: [OWNER, ]},
        '{}_verkey_<any>_<any>'.format(NYM):
            {r: [OWNER] for r in valid_roles},
        '{}_services__[VALIDATOR]'.format(NODE):
            {STEWARD: [OWNER, ]},
        # INDY-410 - steward allowed to demote/promote its validator
        '{}_services_[VALIDATOR]_[]'.format(NODE):
            {TRUSTEE: [], STEWARD: [OWNER, ]},
        '{}_services_[]_[VALIDATOR]'.format(NODE):
            {TRUSTEE: [], STEWARD: [OWNER, ]},
        '{}_node_ip_<any>_<any>'.format(NODE):
            {STEWARD: [OWNER, ]},
        '{}_node_port_<any>_<any>'.format(NODE):
            {STEWARD: [OWNER, ]},
        '{}_client_ip_<any>_<any>'.format(NODE):
            {STEWARD: [OWNER, ]},
        '{}_client_port_<any>_<any>'.format(NODE):
            {STEWARD: [OWNER, ]},
        '{}_blskey_<any>_<any>'.format(NODE):
            {STEWARD: [OWNER, ]},
        '{}_action__start'.format(POOL_UPGRADE):
            {TRUSTEE: []},
        '{}_action_start_cancel'.format(POOL_UPGRADE):
            {TRUSTEE: []},
        '{}_action_<any>_<any>'.format(POOL_RESTART):
            {TRUSTEE: []},
        '{}_action_<any>_<any>'.format(POOL_CONFIG):
            {TRUSTEE: []},
        '{}_<any>_<any>_<any>'.format(VALIDATOR_INFO):
            {TRUSTEE: [], STEWARD: []},
        '{}_<any>_<any>_<any>'.format(LEDGERS_FREEZE):
            {TRUSTEE: []},
    }
    return auth_map


class Authoriser:
    ValidRoles = (TRUSTEE, STEWARD, ENDORSER, NETWORK_MONITOR, None)

    auth_map = None

    @staticmethod
    def isValidRole(role) -> bool:
        return role in Authoriser.ValidRoles

    @staticmethod
    def getRoleFromName(roleName) -> bool:
        if not roleName:
            return
        return Roles[roleName].value

    @staticmethod
    def isValidRoleName(roleName) -> bool:
        if not roleName:
            return True

        try:
            Authoriser.getRoleFromName(roleName)
        except KeyError:
            return False

        return True

    @staticmethod
    def authorised(typ, actorRole, field=None, oldVal=None, newVal=None,
                   isActorOwnerOfSubject=None) -> (bool, str):
        if not Authoriser.auth_map:
            Authoriser.auth_map = generate_auth_map(Authoriser.ValidRoles)
        field = field if field is not None else ""
        oldVal = '' if oldVal is None else \
            str(oldVal).replace('"', '').replace("'", '')
        newVal = '' if newVal is None else \
            str(newVal).replace('"', '').replace("'", '')
        key = '_'.join([typ, field, oldVal, newVal])
        if key not in Authoriser.auth_map:
            any_value = '_'.join([typ, field, '<any>', '<any>'])
            if any_value not in Authoriser.auth_map:
                any_field = '_'.join([typ, "<any>", '<any>', '<any>'])
                if any_field not in Authoriser.auth_map:
                    msg = "key '{}' not found in authorized map".format(key)
                    logger.debug(msg)
                    return False, msg
                else:
                    key = any_field
            else:
                key = any_value
        roles = Authoriser.auth_map[key]
        if actorRole not in roles:
            roles_as_str = [Roles.nameFromValue(role) for role in roles.keys()]
            return False, '{} not in allowed roles {}'. \
                format(Roles.nameFromValue(actorRole), roles_as_str)
        roleDetails = roles[actorRole]
        if len(roleDetails) == 0:
            return True, ''
        else:
            r = OWNER in roleDetails and isActorOwnerOfSubject
            msg = '' if r else 'Only owner is allowed'
            return r, msg



import pytest
import os
import datetime

from common.version import SemVerReleaseVersion

import indy_common
from indy_node.server.upgrade_log import UpgradeLogData, UpgradeLog


@pytest.fixture
def log_file_path(tdir, request):
    return os.path.join(
        tdir,
        "{}.upgrade_log".format(os.path.basename(request.node.nodeid))
    )


@pytest.fixture
def src_version_cls_patched(monkeypatch):
    monkeypatch.setattr(
        indy_common.version,
        'src_version_cls',
        lambda *_: SemVerReleaseVersion
    )


def test_upgrade_log_data_unpack_invalid_version():
    with pytest.raises(TypeError) as excinfo:
        UpgradeLogData(str(datetime.datetime.utcnow()), 123, 'some_id', 'some_pkg')
    assert "'version' should be 'SourceVersion' or 'str'" in str(excinfo.value)


def test_upgrade_log_data_pack_unpack():
    delimiter = '|'
    data = UpgradeLogData(datetime.datetime.utcnow(), '1.2.3', 'some_id', 'some_pkg')
    assert data == UpgradeLogData.unpack(
        data.pack(delimiter=delimiter), delimiter=delimiter
    )


# TODO actually it is already well tested in base calss ActionLog
@pytest.mark.parametrize('ev_type', UpgradeLog.Events)
def test_upgrade_log_append_api(log_file_path, ev_type):
    upgrade_log = UpgradeLog(log_file_path)
    ev_data = UpgradeLogData(datetime.datetime.utcnow(), '1.2.3', 'some_id', 'some_pkg')
    getattr(upgrade_log, "append_{}".format(ev_type.name))(ev_data)
    assert upgrade_log.last_event.data == ev_data
    assert upgrade_log.last_event.ev_type == ev_type


def test_upgrade_log_loads_legacy_data(monkeypatch, log_file_path):

    ev_index = None
    tss = [
        '2019-02-28 07:36:23.135789',
        '2019-02-28 07:37:11.008484',
        '2019-02-28 07:38:33.721644'
    ]

    class datetime_wrapper(datetime.datetime):
        def utcnow():
            return tss[ev_index]

    monkeypatch.setattr(datetime, 'datetime', datetime_wrapper)

    legacy_logs = (
        "{}\tscheduled\t2019-02-28 07:37:11+00:00\t1.6.83\t15513393820971606221\r\n".format(tss[0]) +
        "{}\tstarted\t2019-02-28 07:37:11+00:00\t1.6.83\t15513393820971606221\tindy-node\r\n".format(tss[1]) +
        "{}\tsucceeded\t2019-02-28 07:37:11+00:00\t1.6.83\t15513393820971606221\tindy-node\r\n".format(tss[2])
    )

    new_logs = (
        "{}\tscheduled\t2019-02-28 07:37:11+00:00\t1.6.83\t15513393820971606221\tindy-node\r\n".format(tss[0]) +
        "{}\tstarted\t2019-02-28 07:37:11+00:00\t1.6.83\t15513393820971606221\tindy-node\r\n".format(tss[1]) +
        "{}\tsucceeded\t2019-02-28 07:37:11+00:00\t1.6.83\t15513393820971606221\tindy-node\r\n".format(tss[2])
    )


    with open(log_file_path, 'w', newline='') as f:
        f.write(legacy_logs)
    upgrade_log_legacy = UpgradeLog(log_file_path)

    log_file_path_new = log_file_path + '_new'
    upgrade_log_new = UpgradeLog(log_file_path_new)

    for ev_index, ev in enumerate(upgrade_log_legacy):
        getattr(upgrade_log_new, 'append_' + ev.ev_type.name)(ev.data)

    with open(log_file_path_new, 'r', newline='') as f:
        new_logs_read = f.read()

    assert new_logs_read == new_logs



import random
from indy_node.test.auth_rule.auth_framework.node_services import EditNodeTest
from plenum.common.keygen_utils import init_bls_keys


class EditNodePropertiesTest(EditNodeTest):
    def __init__(self, env, action_id):
        self.node_data = None
        self.node_name = None
        self.steward_wallet = None
        super().__init__(env, action_id)

    def prepare(self):
        req, node_data, node_name, wallet = self._add_node()
        self.send_and_check(req, wallet)
        self.node_data = node_data
        self.node_name = node_name
        self.steward_wallet = wallet

        self.node_req_for_new_rule = self.get_node_req(self.trustee_wallet)
        self.node_req_1 = self.get_node_req()
        self.node_req_2 = self.get_node_req()

        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()

    def get_node_req(self, wallet=None):
        if None in [self.node_data, self.node_name, self.steward_wallet]:
            assert False, "Create a node before changes."
        if wallet is None:
            wallet = self.steward_wallet
        node_data = self._add_changes(self.node_data)
        req, _, _ = self._build_node(wallet,
                                     self.tconf,
                                     self.tdir,
                                     node_name=self.node_name,
                                     node_data=node_data)
        return req, wallet

    def _add_changes(self, node_data):
        pass


class EditNodeIpTest(EditNodePropertiesTest):
    def __init__(self, env, action_id):
        super().__init__(env,
                         action_id)

    def _add_changes(self, node_data):
        sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof = node_data
        nodeIp = "127.0.0.{}".format(random.randint(1, 254))
        return sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof


class EditNodePortTest(EditNodePropertiesTest):
    def __init__(self, env, action_id):
        super().__init__(env,
                         action_id)

    def _add_changes(self, node_data):
        sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof = node_data
        nodePort = random.randint(1000, 9999)
        return sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof


class EditNodeClientIpTest(EditNodePropertiesTest):
    def __init__(self, env, action_id):
        super().__init__(env,
                         action_id)

    def _add_changes(self, node_data):
        sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof = node_data
        clientIp = "127.0.0.{}".format(random.randint(1, 254))
        return sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof


class EditNodeClientPortTest(EditNodePropertiesTest):
    def __init__(self, env, action_id):
        super().__init__(env,
                         action_id)

    def _add_changes(self, node_data):
        sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof = node_data
        clientPort = random.randint(1000, 9999)
        return sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof


class EditNodeBlsTest(EditNodePropertiesTest):
    def __init__(self, env, action_id):
        super().__init__(env,
                         action_id)

    def _add_changes(self, node_data):
        sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof = node_data
        new_blspk, key_proof = init_bls_keys(self.tdir, "node_name")
        bls_key = new_blspk
        return sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof



from collections import OrderedDict
from indy_common.types import ClientGetRevocRegField
from plenum.common.messages.fields import NonEmptyStringField, ConstantField, IntegerField
from indy_common.constants import TXN_TYPE, REVOC_REG_DEF_ID, TIMESTAMP, GET_REVOC_REG


EXPECTED_GET_REVOC_REG_FIELD = OrderedDict([
    (REVOC_REG_DEF_ID, NonEmptyStringField),
    (TIMESTAMP, IntegerField),
    (TXN_TYPE, ConstantField(GET_REVOC_REG))
])


def test_get_revoc_reg_def_schema():
    actual_field_name = OrderedDict(ClientGetRevocRegField.schema).keys()
    assert EXPECTED_GET_REVOC_REG_FIELD.keys() == actual_field_name



import os
from collections import OrderedDict

import pytest

from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER, AuthConstraintOr
from indy_common.test.auth.metadata.helper import set_auth_constraint, PLUGIN_FIELD, build_req_and_action, Action
from plenum.common.constants import TRUSTEE, STEWARD
from plenum.common.exceptions import UnauthorizedClientRequest

MAX_SIG_COUNT = 3


def test_plugin_simple_error_msg_no_plugin_field(write_auth_req_validator):
    set_auth_constraint(write_auth_req_validator,
                        AuthConstraint(role=IDENTITY_OWNER, sig_count=1, need_to_be_owner=True,
                                       metadata={PLUGIN_FIELD: 2}))
    req, actions = build_req_and_action(Action(author=IDENTITY_OWNER, endorser=None,
                                               sigs={IDENTITY_OWNER: 1},
                                               is_owner=True,
                                               amount=None,
                                               extra_sigs=False))

    with pytest.raises(UnauthorizedClientRequest) as excinfo:
        write_auth_req_validator.validate(req, actions)
    assert ("missing required plugin field") in str(excinfo.value)


def test_plugin_simple_error_msg_extra_plugin_field(write_auth_req_validator):
    set_auth_constraint(write_auth_req_validator,
                        AuthConstraint(role=IDENTITY_OWNER, sig_count=1, need_to_be_owner=True))
    req, actions = build_req_and_action(Action(author=IDENTITY_OWNER, endorser=None,
                                               sigs={IDENTITY_OWNER: 1},
                                               is_owner=True,
                                               amount=5,
                                               extra_sigs=False))

    with pytest.raises(UnauthorizedClientRequest) as excinfo:
        write_auth_req_validator.validate(req, actions)
    assert ("plugin field must be absent") in str(excinfo.value)


def test_plugin_simple_error_msg_not_enough_amount(write_auth_req_validator):
    set_auth_constraint(write_auth_req_validator,
                        AuthConstraint(role=IDENTITY_OWNER, sig_count=1, need_to_be_owner=True,
                                       metadata={PLUGIN_FIELD: 10}))
    req, actions = build_req_and_action(Action(author=IDENTITY_OWNER, endorser=None,
                                               sigs={IDENTITY_OWNER: 1},
                                               is_owner=True,
                                               amount=5,
                                               extra_sigs=False))

    with pytest.raises(UnauthorizedClientRequest) as excinfo:
        write_auth_req_validator.validate(req, actions)
    assert ("not enough amount in plugin field") in str(excinfo.value)


def test_plugin_or_error_msg_not_enough_amount(write_auth_req_validator):
    set_auth_constraint(write_auth_req_validator,
                        AuthConstraintOr(auth_constraints=[
                            AuthConstraint(role=TRUSTEE, sig_count=1, need_to_be_owner=False),
                            AuthConstraint(role=STEWARD, sig_count=1, need_to_be_owner=False,
                                           metadata={PLUGIN_FIELD: 10}),
                        ]))

    req, actions = build_req_and_action(Action(author=STEWARD, endorser=None,
                                               sigs={STEWARD: 1},
                                               is_owner=True,
                                               amount=5,
                                               extra_sigs=False))

    with pytest.raises(UnauthorizedClientRequest) as excinfo:
        write_auth_req_validator.validate(req, actions)
    expected = os.linesep.join([
        "Rule for this action is: 1 TRUSTEE signature is required OR 1 STEWARD signature is required with additional metadata new_field 10",
        "Failed checks:",
        "Constraint: 1 TRUSTEE signature is required, Error: Not enough TRUSTEE signatures",
        "Constraint: 1 STEWARD signature is required with additional metadata new_field 10, Error: not enough amount in plugin field"
    ])
    assert expected in str(excinfo.value.reason)


def test_plugin_or_error_msg_not_enough_amount_multiple_metadata_fields(write_auth_req_validator):
    set_auth_constraint(write_auth_req_validator,
                        AuthConstraintOr(auth_constraints=[
                            AuthConstraint(role=TRUSTEE, sig_count=1, need_to_be_owner=False),
                            AuthConstraint(role=STEWARD, sig_count=1, need_to_be_owner=False,
                                           metadata=OrderedDict([
                                               (PLUGIN_FIELD, 10),
                                               ("aaa", "bbb")
                                           ]))
                        ]))
    req, actions = build_req_and_action(Action(author=STEWARD, endorser=None,
                                               sigs={STEWARD: 1},
                                               is_owner=True,
                                               amount=5,
                                               extra_sigs=False))

    with pytest.raises(UnauthorizedClientRequest) as excinfo:
        write_auth_req_validator.validate(req, actions)
    expected = os.linesep.join([
        "Rule for this action is: 1 TRUSTEE signature is required OR 1 STEWARD signature is required with additional metadata new_field 10 aaa bbb",
        "Failed checks:",
        "Constraint: 1 TRUSTEE signature is required, Error: Not enough TRUSTEE signatures",
        "Constraint: 1 STEWARD signature is required with additional metadata new_field 10 aaa bbb, Error: not enough amount in plugin field"
    ])
    assert expected in str(excinfo.value.reason)



import json

import pytest

from indy_common.authorize.auth_constraints import AuthConstraintForbidden
from indy_common.config import SCHEMA_ATTRIBUTES_LIMIT
from indy_common.constants import SCHEMA_NAME, SCHEMA_VERSION, SCHEMA_ATTR_NAMES
from indy_common.types import SchemaField
from indy_node.test.api.helper import validate_write_reply, sdk_write_schema_and_check
from plenum.common.exceptions import RequestRejectedException, RequestNackedException
from plenum.common.util import randomString
from plenum.config import NAME_FIELD_LIMIT


def test_send_schema_multiple_attrib(looper, sdk_pool_handle,
                                     sdk_wallet_endorser):
    sdk_write_schema_and_check(
        looper, sdk_pool_handle,
        sdk_wallet_endorser,
        ["attrib1", "attrib2", "attrib3"],
        "faber",
        "1.4"
    )


def test_send_schema_one_attrib(looper, sdk_pool_handle,
                                sdk_wallet_endorser):
    sdk_write_schema_and_check(
        looper, sdk_pool_handle,
        sdk_wallet_endorser,
        ["attrib1"],
        "University of Saber",
        "1.0"
    )


def test_can_not_send_same_schema(looper, sdk_pool_handle,
                                  sdk_wallet_endorser):
    sdk_write_schema_and_check(
        looper, sdk_pool_handle,
        sdk_wallet_endorser,
        ["attrib1", "attrib2", "attrib3"],
        "business",
        "1.8"
    )

    with pytest.raises(RequestRejectedException,
                       match=str(AuthConstraintForbidden())):
        resp = sdk_write_schema_and_check(
            looper, sdk_pool_handle,
            sdk_wallet_endorser,
            ["attrib1", "attrib2", "attrib3"],
            "business",
            "1.8"
        )
        validate_write_reply(resp)


def test_schema_maximum_attrib(looper, sdk_pool_handle,
                               sdk_wallet_endorser):
    attribs = []
    for i in range(SCHEMA_ATTRIBUTES_LIMIT):
        attribs.append(randomString(NAME_FIELD_LIMIT))

    sdk_write_schema_and_check(
        looper, sdk_pool_handle,
        sdk_wallet_endorser,
        attribs,
        "business1",
        "1.9"
    )


def test_schema_over_maximum_attrib():
    attribs = []
    for i in range(SCHEMA_ATTRIBUTES_LIMIT + 1):
        attribs.append('attrib' + str(i))

    schema = SchemaField()
    with pytest.raises(Exception) as ex_info:
        schema.validate({SCHEMA_NAME: "business2",
                         SCHEMA_VERSION: "2.0",
                         SCHEMA_ATTR_NAMES: attribs})
    ex_info.match(
        "length should be at most {}".format(SCHEMA_ATTRIBUTES_LIMIT)
    )



import pytest
import time

from common.serializers.serialization import domain_state_serializer
from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit
from indy_common.authorize.auth_constraints import ConstraintsSerializer
from indy_common.authorize.auth_map import auth_map
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.test.constants import IDENTIFIERS
from indy_common.types import Request
from indy_node.persistence.idr_cache import IdrCache
from indy_node.server.node import Node
from plenum.common.constants import STEWARD, TRUSTEE

from indy_common.constants import ENDORSER, LOCAL_AUTH_POLICY, NETWORK_MONITOR, CONFIG_LEDGER_AUTH_POLICY
from plenum.common.exceptions import UnauthorizedClientRequest
from plenum.test.helper import randomOperation
from plenum.test.testing_utils import FakeSomething
from state.pruning_state import PruningState
from storage.kv_in_memory import KeyValueStorageInMemory
from indy_node.test.conftest import write_request_validation, write_auth_req_validator, idr_cache


@pytest.fixture(scope='function', params=[True, False])
def is_owner(request):
    return request.param


@pytest.fixture(scope='function', params=[True, False])
def off_ledger_signature(request):
    return request.param


@pytest.fixture(scope='function')
def action_add():
    return AuthActionAdd(txn_type='SomeType',
                         field='some_field',
                         value='new_value')


@pytest.fixture(scope='function')
def action_edit():
    return AuthActionEdit(txn_type='SomeType',
                          field='some_field',
                          old_value='old_value',
                          new_value='new_value')


@pytest.fixture(scope="module")
def constraint_serializer():
    return ConstraintsSerializer(domain_state_serializer)


@pytest.fixture(scope="module")
def config_state(constraint_serializer):
    state = PruningState(KeyValueStorageInMemory())
    Node.add_auth_rules_to_config_state(state=state,
                                        auth_map=auth_map,
                                        serializer=constraint_serializer)
    return state


@pytest.fixture(scope='module', params=[v[0] for v in IDENTIFIERS.values()])
def identifier(request):
    return request.param


@pytest.fixture(scope='module')
def req(identifier):
    return Request(identifier=identifier,
                   operation=randomOperation(),
                   signature='signature')



import pytest

from indy_node.test.upgrade.conftest import EXT_PKT_NAME, EXT_PKT_VERSION
from plenum.common.exceptions import RequestNackedException
from plenum.test.pool_transactions.helper import sdk_add_new_nym

from indy_node.test.upgrade.helper import sdk_ensure_upgrade_sent


@pytest.fixture
def pckg(request):
    return EXT_PKT_NAME, EXT_PKT_VERSION


def test_forced_request_validation(looper, txnPoolNodeSet,
                                   sdk_pool_handle, sdk_wallet_steward,
                                   sdk_wallet_trustee, validUpgradeExpForceTrue):
    for node in txnPoolNodeSet[2:]:
        node.upgrader.check_upgrade_possible = lambda a, b, c: None
    for node in txnPoolNodeSet[:2]:
        node.upgrader.check_upgrade_possible = lambda a, b, c: 'some exception'

    assert all(n.spylog.getAll('processPropagate') == [] for n in txnPoolNodeSet)

    with pytest.raises(RequestNackedException, match='some exception'):
        sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, validUpgradeExpForceTrue)

    assert all(len(n.spylog.getAll('processPropagate')) == 2 for n in txnPoolNodeSet[:2])
    assert all(len(n.spylog.getAll('processPropagate')) == 1 for n in txnPoolNodeSet[2:])

    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_steward)



import json
from random import randint

import pytest
from plenum.common.exceptions import RequestNackedException
from plenum.test.helper import sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request

from plenum.common.util import randomString
from plenum.common.constants import TXN_TYPE, GET_TXN_AUTHOR_AGREEMENT_AML, CURRENT_PROTOCOL_VERSION
from plenum.common.types import OPERATION, f


def test_get_taa_aml_static_validation_fails(looper, txnPoolNodeSet, sdk_pool_handle, sdk_wallet_client):
    req = {
        OPERATION: {
            TXN_TYPE: GET_TXN_AUTHOR_AGREEMENT_AML,
            'timestamp': randint(1, 2147483647),
            'version': randomString()
        },
        f.IDENTIFIER.nm: sdk_wallet_client[1],
        f.REQ_ID.nm: randint(1, 2147483647),
        f.PROTOCOL_VERSION.nm: CURRENT_PROTOCOL_VERSION
    }
    rep = sdk_sign_and_send_prepared_request(looper, sdk_wallet_client, sdk_pool_handle, json.dumps(req))
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [rep])
    e.match('cannot be used in GET_TXN_AUTHOR_AGREEMENT_AML request together')



# Current network
NETWORK_NAME = None

# Disable stdout logging
enableStdOutLogging = False

# Directory to store ledger.
LEDGER_DIR = '/var/lib/indy'

# Directory to store logs.
LOG_DIR = '/var/log/indy'

# Directory to store keys.
KEYS_DIR = '/var/lib/indy'

# Directory to store genesis transactions files.
GENESIS_DIR = '/var/lib/indy'

# Directory to store backups.
BACKUP_DIR = '/var/lib/indy/backup'

# Directory to store plugins.
PLUGINS_DIR = '/var/lib/indy/plugins'

# Directory to store node info.
NODE_INFO_DIR = '/var/lib/indy'



import pytest
import json
import os

from indy_common.node_version import NodeVersion, InvalidVersionError
from plenum.__metadata__ import (
    set_version, load_version,
    set_manifest, load_manifest
)


def tmp_file_path(tdir, request, file_name):
    return os.path.join(
        tdir,
        "{}.{}".format(os.path.basename(request.node.nodeid), file_name)
    )


@pytest.fixture
def version_file_path(tdir, request):
    return tmp_file_path(tdir, request, '__version__.json')


@pytest.fixture
def manifest_file_path(tdir, request):
    return tmp_file_path(tdir, request, '__manifest__.json')


def idfn(v):
    return str(v).replace(' ', '')


@pytest.mark.parametrize(
    'version',
    [
        '1',
        2,
        # (1, 2), TODO uncomment once new release logic becomes completed
        (1, 2, 3, 4, 5),
        (1, 2, 3, 'alpha', 5),
        (1, 2, 3, 'dev', 5, 6)
    ],
    ids=idfn
)
def test_load_version_invalid(version, version_file_path):
    with open(version_file_path, 'w') as _f:
        json.dump(version, _f)
    with pytest.raises(InvalidVersionError):
        load_version(version_file_path)


# TODO ??? wider coverage

@pytest.mark.parametrize(
    'version',
    [
        'a1.2.3',
        '1.2.3a1',
        '1.2.3b2',
        '1.2.3.post1',
        '1!1.2.3',
        '1.2.3+1',
        '1',
        # '1.2', TODO uncomment once new release logic becomes completed
        '1.2.3.4',
        2
    ],
    ids=idfn
)
def test_set_version_invalid(version, version_file_path):
    with pytest.raises(InvalidVersionError):
        set_version(version, version_file_path)


@pytest.mark.parametrize(
    'version',
    [
        '1.2.3',
        '1.2.3.rc1',
        '1.2.3.dev2',
    ]
)
def test_set_load_version_valid(version, version_file_path):
    set_version(version, version_file_path)
    assert load_version(version_file_path) == NodeVersion(version)



def test_load_absent_manifest(manifest_file_path):
    assert load_manifest(manifest_file_path) is None


# TODO ??? wider coverage
@pytest.mark.parametrize(
    'manifest',
    [
        1,
        [1, 2],
        None,
        {'k1': 'v1', "2": 2},
    ],
    ids=idfn
)
def test_set_load_manifest_valid(manifest, manifest_file_path):
    set_manifest(manifest, manifest_file_path)
    assert load_manifest(manifest_file_path) == manifest



import json

from indy_common.constants import CRED_DEF_ID, CLAIM_DEF_SCHEMA_REF, CLAIM_DEF_SIGNATURE_TYPE, CLAIM_DEF_TAG
from indy_common.state.domain import make_state_path_for_claim_def
from plenum.test.helper import sdk_send_and_check
from plenum.test.helper import sdk_sign_request_from_dict


def test_send_revoc_reg_def(looper,
                            txnPoolNodeSet,
                            sdk_wallet_steward,
                            sdk_pool_handle,
                            build_revoc_def_by_default,
                            send_claim_def):
    txns_count_before = set([n.domainLedger.size for n in txnPoolNodeSet])
    assert len(txns_count_before) == 1, "Ledger size for nodes are not equal"
    _, author_did = sdk_wallet_steward
    claim_def_req = send_claim_def[0]
    revoc_req = build_revoc_def_by_default
    revoc_req['operation'][CRED_DEF_ID] = make_state_path_for_claim_def(author_did,
                                                                        str(claim_def_req['operation'][
                                                                                CLAIM_DEF_SCHEMA_REF]),
                                                                        claim_def_req['operation'][
                                                                            CLAIM_DEF_SIGNATURE_TYPE],
                                                                        claim_def_req['operation'][CLAIM_DEF_TAG]
                                                                        ).decode()
    revoc_req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, revoc_req['operation'])
    sdk_send_and_check([json.dumps(revoc_req)], looper, txnPoolNodeSet, sdk_pool_handle)
    txns_count_after = set([n.domainLedger.size for n in txnPoolNodeSet])
    assert len(txns_count_after) == 1, "Ledger size for nodes are not equal"
    # REVOC_REG_DEF transaction was written
    assert txns_count_after.pop() - txns_count_before.pop() == 1



import os
import shutil
import psutil

from stp_core.common.log import getlogger

logger = getlogger()


class TempStorage:
    @staticmethod
    def getOpenFdsUnder(dirPath):
        proc = psutil.Process()
        # return {f.fd for f in proc.open_files()
        #         if os.path.dirname(f.path) == dirPath}
        r = set()
        for f in proc.open_files():
            if os.path.dirname(f.path) == dirPath:
                logger.debug(
                    'Going to close before deleting: {}'.format(f.path))
                r.add(f.fd)
        return r

    @staticmethod
    def closeFdsUnder(dirPath):
        for fd in TempStorage.getOpenFdsUnder(dirPath):
            os.close(fd)

    @staticmethod
    def cleanupDirectory(dirPath):
        try:
            # Uncommenting the next line causes random socket failures, from
            # Python docs, "This function is intended for low-level I/O and
            # must be applied to a file descriptor as returned by
            # os.open() or pipe()"
            # TempStorage.closeFdsUnder(dirPath)
            shutil.rmtree(dirPath)
        except Exception as ex:
            logger.debug(
                "Error while removing temporary directory {}".format(ex))

    def cleanupDataLocation(self):
        self.cleanupDirectory(self.dataLocation)
        try:
            odbClient = self.graphStore.client
            odbClient.db_drop(self.name)
            logger.debug("Dropped db {}".format(self.name))
        except Exception as ex:
            logger.debug("Error while dropping db {}: {}".format(self.name,
                                                                 ex))



from indy_common.authorize.auth_actions import AuthActionAdd
from plenum.common.constants import TXN_AUTHOR_AGREEMENT


def test_auth_txn_author_agreement_with_new_auth_map(write_request_validation, is_owner, req):
    authorized = req.identifier == "trustee_identifier"
    assert authorized == write_request_validation(req,
                                                  [AuthActionAdd(txn_type=TXN_AUTHOR_AGREEMENT,
                                                                 field='some_field',
                                                                 value='some_value',
                                                                 is_owner=is_owner)])



import json

import pytest

from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.authorize.auth_constraints import ROLE, AuthConstraint, AuthConstraintAnd, AuthConstraintOr
from indy_common.constants import NYM
from indy_node.test.auth_rule.helper import create_verkey_did, sdk_send_and_check_auth_rule_request
from indy_node.test.nym_txn.test_nym_additional import get_nym
from plenum.common.constants import STEWARD, STEWARD_STRING, TRUSTEE, TRUSTEE_STRING
from plenum.common.exceptions import RequestRejectedException
from plenum.common.util import randomString
from plenum.test.conftest import sdk_wallet_client
from plenum.test.helper import sdk_get_and_check_replies, sdk_send_signed_requests, sdk_sign_request_objects, \
    sdk_json_to_request_object, sdk_multi_sign_request_objects
from plenum.test.pool_transactions.helper import prepare_nym_request, \
    sdk_sign_and_send_prepared_request, sdk_add_new_nym

auth_constraint = AuthConstraintOr(auth_constraints=[AuthConstraintAnd(auth_constraints=[AuthConstraint(role=STEWARD,
                                                                                                        sig_count=2),
                                                                                         AuthConstraint(role=TRUSTEE,
                                                                                                        sig_count=1)]),
                                                     AuthConstraint(role=TRUSTEE,
                                                                    sig_count=2)
                                                     ])


@pytest.fixture(scope='module', params=[(2, 0), (3, 0), (1, 2), (1, 3), (2, 1)])
def wallets_for_success(request, sdk_wallet_steward_list, sdk_wallet_trustee_list):
    return (
        sdk_wallet_trustee_list[:request.param[0]] +
        sdk_wallet_steward_list[:request.param[1]]
    )


@pytest.fixture(scope='module', params=[(0, 3), (1, 0), (1, 1)])
def wallets_for_fail(request, sdk_wallet_steward_list, sdk_wallet_trustee_list):
    return (
        sdk_wallet_trustee_list[:request.param[0]] +
        sdk_wallet_steward_list[:request.param[1]]
    )


@pytest.fixture(scope='module')
def sdk_wallet_steward_list(looper,
                            sdk_wallet_trustee,
                            sdk_pool_handle):
    sdk_wallet_steward_list = []
    for i in range(3):
        wallet = sdk_add_new_nym(looper,
                                 sdk_pool_handle,
                                 sdk_wallet_trustee,
                                 alias='steward{}'.format(i),
                                 role=STEWARD_STRING)
        sdk_wallet_steward_list.append(wallet)
    return sdk_wallet_steward_list


@pytest.fixture(scope='module')
def prepare_auth_map(looper,
                     sdk_wallet_trustee,
                     sdk_pool_handle):
    wh, _ = sdk_wallet_trustee
    did1, verkey1 = create_verkey_did(looper, wh)
    """Adding new steward for old auth rules"""
    add_new_nym(looper,
                sdk_pool_handle,
                [sdk_wallet_trustee],
                'newSteward1',
                STEWARD_STRING,
                dest=did1, verkey=verkey1)
    sdk_send_and_check_auth_rule_request(looper, sdk_pool_handle, sdk_wallet_trustee,
                                         auth_action=ADD_PREFIX,
                                         auth_type=NYM, field=ROLE, new_value=STEWARD, old_value=None,
                                         constraint=auth_constraint.as_dict)


def add_new_nym(looper, sdk_pool_handle, creators_wallets,
                alias=None, role=None, seed=None,
                dest=None, verkey=None, skipverkey=False, no_wait=False):
    seed = seed or randomString(32)
    alias = alias or randomString(5)
    wh, _ = creators_wallets[0]

    # filling nym request and getting steward did
    # if role == None, we are adding client
    nym_request, new_did = looper.loop.run_until_complete(
        prepare_nym_request(creators_wallets[0], seed,
                            alias, role, dest, verkey, skipverkey))

    # sending request using 'sdk_' functions
    signed_reqs = sdk_multi_sign_request_objects(looper, creators_wallets,
                                                 [sdk_json_to_request_object(
                                                     json.loads(nym_request))])
    request_couple = sdk_send_signed_requests(sdk_pool_handle, signed_reqs)[0]
    if no_wait:
        return request_couple
    # waitng for replies
    sdk_get_and_check_replies(looper, [request_couple])


def test_multi_sig_auth_rule(looper,
                             wallets_for_success,
                             sdk_pool_handle,
                             prepare_auth_map):
    wh, _ = wallets_for_success[0]
    did, verkey = create_verkey_did(looper, wh)
    add_new_nym(looper,
                sdk_pool_handle,
                wallets_for_success,
                'newSteward1',
                STEWARD_STRING,
                dest=did)
    get_nym(looper, sdk_pool_handle, wallets_for_success[0], did)


def test_reject_request_with_multi_sig_auth_rule(looper,
                                                 wallets_for_fail,
                                                 sdk_pool_handle,
                                                 prepare_auth_map):
    with pytest.raises(RequestRejectedException, match="Rule for this action is: {}".format(auth_constraint)):
        """Adding new steward for changed auth rules"""
        add_new_nym(looper,
                    sdk_pool_handle,
                    wallets_for_fail,
                    'newSteward2',
                    STEWARD_STRING)


def test_validate_by_auth_rule_without_signatures(looper,  # noqa: F811
                                                  sdk_wallet_client,
                                                  sdk_pool_handle,
                                                  prepare_auth_map):
    with pytest.raises(RequestRejectedException, match="Rule for this action is: {}".format(auth_constraint)):
        """Adding new steward for changed auth rules"""
        add_new_nym(looper,
                    sdk_pool_handle,
                    [sdk_wallet_client],
                    'newSteward2',
                    STEWARD_STRING)



import pytest
import copy
from indy_common.types import SafeRequest
from indy_common.constants import REVOC_REG_DEF_ID
from indy_common.constants import FROM, TO
from indy_common.types import Request
from indy_node.test.anon_creds.helper import build_get_revoc_reg_delta
from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.exceptions import InvalidClientRequest


def test_revoc_reg_delta_schema_validation_wrong_type(looper, sdk_wallet_steward):
    req = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    req['operation'][FROM] = "10"
    with pytest.raises(TypeError, match="expected types 'int', got 'str'"):
        SafeRequest(**req)


def test_revoc_reg_delta_schema_validation_missed_fields(looper, sdk_wallet_steward):
    req = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    del req['operation'][REVOC_REG_DEF_ID]
    with pytest.raises(TypeError, match="missed fields - {}".format(REVOC_REG_DEF_ID)):
        SafeRequest(**req)


def test_revoc_reg_delta_from_greater_then_to(create_node_and_not_start,
                                              looper, sdk_wallet_steward):
    node = create_node_and_not_start
    req = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    req['operation'][FROM] = 100
    req['operation'][TO] = 20
    with pytest.raises(InvalidClientRequest, match="Timestamp FROM more then TO"):
        node.read_manager.static_validation(Request(**req))



from indy_common.authorize.auth_actions import AuthActionAdd
from indy_common.constants import VALIDATOR_INFO


def test_pool_restart_add_action(write_request_validation, is_owner, req):
    authorized = req.identifier in ("trustee_identifier", "steward_identifier", "network_monitor_identifier")
    assert authorized == write_request_validation(req,
                                                  [AuthActionAdd(txn_type=VALIDATOR_INFO,
                                                                 field='some_field',
                                                                 value='some_value',
                                                                 is_owner=is_owner)])



from copy import deepcopy

from plenum.common.constants import NAME, VERSION, STEWARD_STRING
from plenum.common.exceptions import RequestNackedException, RequestRejectedException
from indy_common.constants import CANCEL, \
    ACTION
from indy_node.test.upgrade.helper import bumpedVersion, sdk_send_upgrade
from plenum.test.helper import sdk_get_bad_response

whitelist = ['Failed to upgrade node']


def testNodeRejectsPoolUpgrade(looper, nodeSet, tdir, sdk_pool_handle,
                               sdk_wallet_trustee, invalidUpgrade):
    req = sdk_send_upgrade(looper, sdk_pool_handle, sdk_wallet_trustee, invalidUpgrade)
    sdk_get_bad_response(looper, [req], RequestNackedException, 'since time span between upgrades')


def testOnlyTrusteeCanSendPoolUpgrade(looper, sdk_pool_handle, sdk_wallet_steward, validUpgrade):
    # A steward sending POOL_UPGRADE but txn fails
    validUpgrade = deepcopy(validUpgrade)
    validUpgrade[NAME] = 'upgrade-20'
    validUpgrade[VERSION] = bumpedVersion(validUpgrade['version'])
    req = sdk_send_upgrade(looper, sdk_pool_handle, sdk_wallet_steward, validUpgrade)
    sdk_get_bad_response(looper, [req], RequestRejectedException, 'Not enough TRUSTEE signatures')


def testNonTrustyCannotCancelUpgrade(looper, validUpgradeSent, sdk_pool_handle,
                                     sdk_wallet_steward, validUpgrade):
    validUpgradeCopy = deepcopy(validUpgrade)
    validUpgradeCopy[ACTION] = CANCEL
    req = sdk_send_upgrade(looper, sdk_pool_handle, sdk_wallet_steward, validUpgradeCopy)
    sdk_get_bad_response(looper, [req], RequestRejectedException, 'Not enough TRUSTEE signatures')


def test_accept_then_reject_upgrade(
        looper, sdk_pool_handle, sdk_wallet_trustee, validUpgradeSent, validUpgrade):
    upgrade_name = validUpgrade[NAME]
    error_msg = "Upgrade '{}' is already scheduled".format(upgrade_name)

    validUpgrade2 = deepcopy(validUpgrade)

    req = sdk_send_upgrade(looper, sdk_pool_handle, sdk_wallet_trustee, validUpgrade2)
    sdk_get_bad_response(looper, [req], RequestRejectedException, error_msg)


def testOnlyTrusteeCanSendPoolUpgradeForceTrue(
        looper, sdk_pool_handle, sdk_wallet_steward, validUpgradeExpForceTrue):
    req = sdk_send_upgrade(looper, sdk_pool_handle, sdk_wallet_steward, validUpgradeExpForceTrue)
    sdk_get_bad_response(looper, [req], RequestNackedException, 'Not enough TRUSTEE signatures')



import pytest

from plenum.common.exceptions import RequestRejectedException
from indy_node.test.suspension.helper import sdk_suspend_role
from plenum.common.util import randomString
from plenum.test.pool_transactions.helper import demote_node, sdk_add_new_nym
from stp_core.loop.eventually import eventually
from stp_core.common.log import getlogger
from plenum.common.constants import STEWARD_STRING, TRUSTEE_STRING
from plenum.test.pool_transactions.test_suspend_node import \
    checkNodeNotInNodeReg
from indy_common.constants import ENDORSER_STRING

logger = getlogger()


@pytest.fixture(scope="module")
def another_trustee(looper, nodeSet, sdk_pool_handle, sdk_wallet_trustee):
    return sdk_add_new_nym(looper, sdk_pool_handle,
                           sdk_wallet_trustee, 'newTrustee', TRUSTEE_STRING)


@pytest.fixture(scope="module")
def another_steward(looper, nodeSet, sdk_pool_handle, sdk_wallet_trustee):
    return sdk_add_new_nym(looper, sdk_pool_handle,
                           sdk_wallet_trustee, 'newSteward', STEWARD_STRING)


@pytest.fixture(scope="module")
def another_steward1(looper, nodeSet, sdk_pool_handle, sdk_wallet_trustee):
    return sdk_add_new_nym(looper, sdk_pool_handle,
                           sdk_wallet_trustee, 'newSteward1', STEWARD_STRING)


@pytest.fixture(scope="module")
def another_endorser(looper, nodeSet, sdk_pool_handle, sdk_wallet_trustee):
    return sdk_add_new_nym(looper, sdk_pool_handle,
                           sdk_wallet_trustee, 'newEndorser', ENDORSER_STRING)


@pytest.fixture(scope="module")
def another_endorser1(looper, nodeSet, sdk_pool_handle, sdk_wallet_trustee):
    return sdk_add_new_nym(looper, sdk_pool_handle,
                           sdk_wallet_trustee, 'newEndorser1', ENDORSER_STRING)


def testTrusteeAddingAnotherTrustee(another_trustee):
    pass


def testTrusteeAddingSteward(looper, sdk_pool_handle, another_steward):
    # The new Steward adds a ENDORSER
    sdk_add_new_nym(looper, sdk_pool_handle, another_steward, role=ENDORSER_STRING)


def testTrusteeAddingEndorser(looper, sdk_pool_handle, another_endorser):
    # The new TEndorser adds a NYM
    sdk_add_new_nym(looper, sdk_pool_handle, another_endorser)


def testStewardSuspensionByTrustee(looper, sdk_pool_handle,
                                   another_trustee, another_steward):
    _, did_stew = another_steward
    sdk_suspend_role(looper, sdk_pool_handle, another_trustee, did_stew)
    with pytest.raises(RequestRejectedException):
        sdk_add_new_nym(looper, sdk_pool_handle,
                        another_steward, role=ENDORSER_STRING)


def testEndorserSuspensionByTrustee(
        looper, sdk_pool_handle, another_trustee, another_endorser):
    _, did_ta = another_endorser
    sdk_suspend_role(looper, sdk_pool_handle, another_trustee, did_ta)
    with pytest.raises(RequestRejectedException):
        sdk_add_new_nym(looper, sdk_pool_handle,
                        another_endorser, alias=randomString())


def testTrusteeSuspensionByTrustee(looper, sdk_pool_handle, sdk_wallet_trustee,
                                   another_trustee, another_steward1):
    # trustee suspension by trustee is succeed
    _, did = another_trustee
    sdk_suspend_role(looper, sdk_pool_handle, sdk_wallet_trustee, did)

    # trustee suspension by steward1 is failed
    _, did = sdk_wallet_trustee
    with pytest.raises(RequestRejectedException) as e:
        sdk_suspend_role(looper, sdk_pool_handle, another_steward1, did)
    e.match("Not enough TRUSTEE signatures")


# Keep the test below at the end of the suite since it will make one of the
# nodes inactive, unless you are planning to add new nodes.
def testValidatorSuspensionByTrustee(sdk_wallet_trustee, sdk_pool_handle, looper, nodeSet):
    node = nodeSet[-1]
    demote_node(looper, sdk_wallet_trustee, sdk_pool_handle, node)
    for n in nodeSet[:-1]:
        looper.run(eventually(checkNodeNotInNodeReg, n, node.name))



import pytest

from indy_common.state import config
from indy_common.types import AuthRuleField
from indy_node.server.request_handlers.config_req_handlers.auth_rule.static_auth_rule_helper import StaticAuthRuleHelper
from plenum.common.types import OPERATION
from plenum.common.constants import TXN_TYPE, TRUSTEE, STEWARD, DATA, STATE_PROOF
from plenum.common.exceptions import RequestNackedException

from indy_common.authorize.auth_actions import ADD_PREFIX, EDIT_PREFIX
from indy_common.authorize.auth_constraints import ROLE, AuthConstraintForbidden
from indy_common.authorize.auth_map import auth_map
from indy_common.constants import NYM, ENDORSER, AUTH_ACTION, AUTH_TYPE, FIELD, NEW_VALUE, \
    OLD_VALUE, SCHEMA, CONSTRAINT, AUTH_RULE

from plenum.test.helper import (
    sdk_gen_request, sdk_sign_and_submit_req, sdk_get_and_check_replies,
    sdk_sign_and_submit_req_obj
)
from indy_node.test.helper import build_auth_rule_request_json, generate_constraint_entity
from indy_node.test.auth_rule.helper import generate_constraint_list, \
    sdk_send_and_check_auth_rule_request, generate_key, sdk_send_and_check_get_auth_rule_request, \
    sdk_send_and_check_get_auth_rule_invalid_request


RESULT = "result"


def test_fail_get_auth_rule_with_incorrect_key(looper,
                                               sdk_wallet_trustee,
                                               sdk_pool_handle):
    key = generate_key()
    key[AUTH_TYPE] = "wrong_txn_type"
    with pytest.raises(RequestNackedException, match="Unknown authorization rule: key .* "
                                                     "is not found in authorization map."):
        sdk_send_and_check_get_auth_rule_invalid_request(
            looper, sdk_pool_handle, sdk_wallet_trustee, **key
        )[0]

    del key[AUTH_TYPE]
    with pytest.raises(RequestNackedException, match="Not enough fields to build an auth key."):
        sdk_send_and_check_get_auth_rule_invalid_request(
            looper, sdk_pool_handle, sdk_wallet_trustee, **key
        )[0]


def _check_key(key, resp_key):
    assert key[AUTH_ACTION] in resp_key[AUTH_ACTION]
    assert key[AUTH_TYPE] in resp_key[AUTH_TYPE]
    assert key[FIELD] in resp_key[FIELD]
    assert key[NEW_VALUE] in resp_key[NEW_VALUE]
    if key[AUTH_ACTION] == EDIT_PREFIX:
        assert key[OLD_VALUE] in resp_key[OLD_VALUE]
    else:
        assert OLD_VALUE not in resp_key


def test_get_one_auth_rule_transaction(looper,
                                       sdk_wallet_trustee,
                                       sdk_pool_handle):
    key = generate_key()
    str_key = StaticAuthRuleHelper.get_auth_key(key)
    req, resp = sdk_send_and_check_get_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet_trustee, **key
    )[0]

    for resp_rule in resp[RESULT][DATA]:
        _check_key(key, resp_rule)
        assert auth_map.get(str_key).as_dict == resp_rule[CONSTRAINT]


def test_get_unknown_auth_rule_transaction_is_rejected(
    looper, sdk_wallet_trustee, sdk_pool_handle
):
    key = generate_key(auth_action=ADD_PREFIX, auth_type=NYM,
                       field=ROLE, new_value='*')
    with pytest.raises(
        RequestNackedException,
        match=r"Unknown authorization rule: key '1--ADD--role--\*--\*' "
              "is not found in authorization map"
    ):
        sdk_send_and_check_get_auth_rule_request(
            looper, sdk_pool_handle, sdk_wallet_trustee, **key
        )


def test_get_one_disabled_auth_rule_transaction(looper,
                                                sdk_wallet_trustee,
                                                sdk_pool_handle):
    key = generate_key(auth_action=EDIT_PREFIX, auth_type=SCHEMA,
                       field='*', old_value='*', new_value='*')
    req, resp = sdk_send_and_check_get_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet_trustee, **key
    )[0]

    result = resp["result"][DATA]
    assert len(result) == 1
    _check_key(key, result[0])
    assert AuthConstraintForbidden().as_dict == result[0][CONSTRAINT]


def test_get_all_auth_rule_transactions(looper,
                                        sdk_wallet_trustee,
                                        sdk_pool_handle):
    resp = sdk_send_and_check_get_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet_trustee
    )

    result = resp[0][1]["result"][DATA]
    for i, (auth_key, constraint) in enumerate(auth_map.items()):
        rule = result[i]
        assert auth_key == StaticAuthRuleHelper.get_auth_key(rule)
        if constraint is None:
            assert {} == rule[CONSTRAINT]
        else:
            assert constraint.as_dict == rule[CONSTRAINT]


def test_get_one_auth_rule_transaction_after_write(looper,
                                                   sdk_wallet_trustee,
                                                   sdk_pool_handle):
    auth_action = ADD_PREFIX
    auth_type = NYM
    field = ROLE
    new_value = ENDORSER
    constraint = generate_constraint_list(auth_constraints=[generate_constraint_entity(role=TRUSTEE),
                                                            generate_constraint_entity(role=STEWARD)])
    resp = sdk_send_and_check_auth_rule_request(looper,
                                                sdk_pool_handle,
                                                sdk_wallet_trustee,
                                                auth_action=auth_action, auth_type=auth_type,
                                                field=field, new_value=new_value,
                                                constraint=constraint)
    dict_auth_key = generate_key(auth_action=auth_action, auth_type=auth_type,
                                 field=field, new_value=new_value)
    resp = sdk_send_and_check_get_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet_trustee, **dict_auth_key
    )
    result = resp[0][1]["result"][DATA]
    assert len(result) == 1
    _check_key(dict_auth_key, result[0])
    assert constraint == result[0][CONSTRAINT]
    assert resp[0][1]["result"][STATE_PROOF]


def test_get_all_auth_rule_transactions_after_write(looper,
                                                    sdk_wallet_trustee,
                                                    sdk_pool_handle):
    auth_action = ADD_PREFIX
    auth_type = NYM
    field = ROLE
    new_value = ENDORSER
    constraint = generate_constraint_list(auth_constraints=[generate_constraint_entity(role=TRUSTEE),
                                                            generate_constraint_entity(role=STEWARD)])
    resp = sdk_send_and_check_auth_rule_request(looper,
                                                sdk_pool_handle,
                                                sdk_wallet_trustee,
                                                auth_action=auth_action,
                                                auth_type=auth_type,
                                                field=field,
                                                new_value=new_value,
                                                constraint=constraint)
    str_auth_key = StaticAuthRuleHelper.get_auth_key(resp[0][0][OPERATION])
    resp = sdk_send_and_check_get_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet_trustee
    )

    result = resp[0][1]["result"][DATA]
    for rule in result:
        key = StaticAuthRuleHelper.get_auth_key(rule)
        if key == str_auth_key:
            assert constraint == rule[CONSTRAINT]
        else:
            assert auth_map[key].as_dict == rule[CONSTRAINT]


def test_auth_rule_after_get_auth_rule_as_is(
    looper, sdk_pool_handle, sdk_wallet_trustee
):
    # get all auth rules
    auth_rules_resp = sdk_send_and_check_get_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet_trustee
    )
    auth_rules = auth_rules_resp[0][1][RESULT][DATA]

    for rule in auth_rules:
        # prepare action key
        dict_key = dict(rule)
        dict_key.pop(CONSTRAINT)

        # prepare "operation" to send AUTH_RULE txn
        op = dict(rule)
        op[TXN_TYPE] = AUTH_RULE

        # send AUTH_RULE txn
        req_obj = sdk_gen_request(op, identifier=sdk_wallet_trustee[1])
        req = sdk_sign_and_submit_req_obj(looper,
                                          sdk_pool_handle,
                                          sdk_wallet_trustee,
                                          req_obj)
        sdk_get_and_check_replies(looper, [req])

        # send GET_AUTH_RULE
        get_response = sdk_send_and_check_get_auth_rule_request(
            looper, sdk_pool_handle, sdk_wallet_trustee, **dict_key)
        # check response
        result = get_response[0][1]["result"][DATA]
        assert len(result) == 1
        _check_key(dict_key, result[0])
        assert rule[CONSTRAINT] == result[0][CONSTRAINT]
        assert get_response[0][1]["result"][STATE_PROOF]


def test_auth_rule_after_get_auth_rule_as_is_using_sdk(
    looper, sdk_pool_handle, sdk_wallet_trustee
):
    # get all auth rules
    auth_rules_resp = sdk_send_and_check_get_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet_trustee
    )
    auth_rules = auth_rules_resp[0][1][RESULT][DATA]

    for rule in auth_rules:
        # prepare action key
        dict_key = dict(rule)
        dict_key.pop(CONSTRAINT)

        # prepare AUTH_RULE request
        req_json = build_auth_rule_request_json(
            looper, sdk_wallet_trustee[1], **rule
        )

        # send AUTH_RULE txn
        req = sdk_sign_and_submit_req(sdk_pool_handle,
                                      sdk_wallet_trustee,
                                      req_json)
        sdk_get_and_check_replies(looper, [req])

        # send GET_AUTH_RULE
        get_response = sdk_send_and_check_get_auth_rule_request(
            looper, sdk_pool_handle, sdk_wallet_trustee, **dict_key)
        # check response
        result = get_response[0][1]["result"][DATA]
        assert len(result) == 1
        _check_key(dict_key, result[0])
        assert rule[CONSTRAINT] == result[0][CONSTRAINT]
        assert get_response[0][1]["result"][STATE_PROOF]


def test_auth_rule_after_get_auth_rule_as_is_except_constraint(
    looper, sdk_wallet_trustee, sdk_pool_handle
):
    constraint = generate_constraint_list(auth_constraints=[generate_constraint_entity(role=TRUSTEE),
                                                            generate_constraint_entity(role=STEWARD)])
    # get all auth rules
    auth_rules_resp = sdk_send_and_check_get_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet_trustee
    )
    auth_rules = auth_rules_resp[0][1][RESULT][DATA]
    rule = auth_rules[0]

    # prepare action key
    dict_key = dict(rule)
    dict_key.pop(CONSTRAINT)

    # prepare "operation" to send AUTH_RULE txn
    op = dict(rule)
    op[TXN_TYPE] = AUTH_RULE
    op[CONSTRAINT] = constraint

    # send AUTH_RULE txn
    req_obj = sdk_gen_request(op, identifier=sdk_wallet_trustee[1])
    req = sdk_sign_and_submit_req_obj(looper,
                                      sdk_pool_handle,
                                      sdk_wallet_trustee,
                                      req_obj)
    sdk_get_and_check_replies(looper, [req])

    # send GET_AUTH_RULE
    get_response = sdk_send_and_check_get_auth_rule_request(
        looper, sdk_pool_handle, sdk_wallet_trustee, **dict_key
    )

    # check response
    result = get_response[0][1]["result"][DATA]
    assert len(result) == 1
    _check_key(dict_key, result[0])
    assert constraint == result[0][CONSTRAINT]
    assert get_response[0][1]["result"][STATE_PROOF]



import json

import pytest
from indy.did import create_and_store_my_did
from indy.ledger import build_nym_request

from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint, OFF_LEDGER_SIGNATURE, CONSTRAINT_ID
from indy_common.constants import CONSTRAINT
from indy_node.test.auth_rule.test_auth_rules_transaction import RESULT
from indy_node.test.helper import build_auth_rule_request_json, sdk_send_and_check_req_json, \
    sdk_send_and_check_get_auth_rule_request, sdk_send_and_check_auth_rule_request
from plenum.common.constants import ROLE, VERKEY, NYM, DATA
from plenum.common.exceptions import RequestNackedException, RequestRejectedException
from plenum.common.util import randomString
from plenum.server.request_handlers.utils import get_nym_details
from plenum.test.helper import sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request

NEW_ROLE = None


# Note: Similar tests written in plenum, because of specific of signature
# validation, which requires to duplicate verification code

@pytest.fixture(scope='function')
def nym_txn_data(looper, sdk_wallet_client):
    seed = randomString(32)

    wh, _ = sdk_wallet_client
    sender_did, sender_verkey = \
        looper.loop.run_until_complete(create_and_store_my_did(wh, json.dumps({'seed': seed})))
    return wh, randomString(5), sender_did, sender_verkey


def change_auth_rule(looper, sdk_pool_handle, sdk_wallet_trustee, constraint):
    req = build_auth_rule_request_json(
        looper, sdk_wallet_trustee[1],
        auth_action=ADD_PREFIX,
        auth_type=NYM,
        field=ROLE,
        old_value=None,
        new_value=NEW_ROLE,
        constraint=constraint.as_dict
    )

    sdk_send_and_check_req_json(looper, sdk_pool_handle, sdk_wallet_trustee, req)


def test_create_did_without_endorser_fails(looper, txnPoolNodeSet, nym_txn_data, sdk_pool_handle):
    wh, alias, sender_did, sender_verkey = nym_txn_data
    nym_request = looper.loop.run_until_complete(
        build_nym_request(sender_did, sender_did, sender_verkey, alias, NEW_ROLE))

    request_couple = sdk_sign_and_send_prepared_request(looper, (wh, sender_did), sdk_pool_handle, nym_request)
    with pytest.raises(RequestRejectedException, match='is not found in the Ledger'):
        sdk_get_and_check_replies(looper, [request_couple])


def test_create_did_without_endorser(looper, txnPoolNodeSet, nym_txn_data, sdk_pool_handle, sdk_wallet_trustee):
    change_auth_rule(looper, sdk_pool_handle, sdk_wallet_trustee, constraint=AuthConstraint(role='*',
                                                                                            sig_count=1,
                                                                                            off_ledger_signature=True))

    wh, alias, sender_did, sender_verkey = nym_txn_data
    nym_request = looper.loop.run_until_complete(
        build_nym_request(sender_did, sender_did, sender_verkey, alias, NEW_ROLE))

    request_couple = sdk_sign_and_send_prepared_request(looper, (wh, sender_did), sdk_pool_handle, nym_request)
    sdk_get_and_check_replies(looper, [request_couple])

    details = get_nym_details(txnPoolNodeSet[0].states[1], sender_did, is_committed=True)
    assert details[ROLE] == NEW_ROLE
    assert details[VERKEY] == sender_verkey


def test_create_did_without_endorser_empty_verkey(looper, nym_txn_data, sdk_wallet_client, sdk_pool_handle):
    wh, alias, sender_did, sender_verkey = nym_txn_data

    nym_request = looper.loop.run_until_complete(build_nym_request(sender_did, sender_did, None, alias, NEW_ROLE))

    request_couple = sdk_sign_and_send_prepared_request(looper, (wh, sender_did), sdk_pool_handle, nym_request)

    with pytest.raises(RequestNackedException, match='verkey for {} cannot be found'.format(sender_did)):
        sdk_get_and_check_replies(looper, [request_couple])


def test_create_did_without_endorser_with_different_dest(looper, nym_txn_data, sdk_wallet_client, sdk_pool_handle):
    wh, alias, sender_did, sender_verkey = nym_txn_data

    nym_request = looper.loop.run_until_complete(
        build_nym_request(sender_did, sdk_wallet_client[1], sender_verkey, alias, NEW_ROLE))

    request_couple = sdk_sign_and_send_prepared_request(looper, (wh, sender_did), sdk_pool_handle, nym_request)

    with pytest.raises(RequestNackedException, match='verkey for {} cannot be found'.format(sender_did)):
        sdk_get_and_check_replies(looper, [request_couple])


def test_create_did_without_endorser_sig_count_0(looper, txnPoolNodeSet, nym_txn_data, sdk_pool_handle,
                                                 sdk_wallet_trustee):
    change_auth_rule(looper, sdk_pool_handle, sdk_wallet_trustee, constraint=AuthConstraint(role='*',
                                                                                            sig_count=0,
                                                                                            off_ledger_signature=True))

    wh, alias, sender_did, sender_verkey = nym_txn_data
    nym_request = looper.loop.run_until_complete(
        build_nym_request(sender_did, sender_did, sender_verkey, alias, NEW_ROLE))

    request_couple = sdk_sign_and_send_prepared_request(looper, (wh, sender_did), sdk_pool_handle, nym_request)
    sdk_get_and_check_replies(looper, [request_couple])

    details = get_nym_details(txnPoolNodeSet[0].states[1], sender_did, is_committed=True)
    assert details[ROLE] == NEW_ROLE
    assert details[VERKEY] == sender_verkey


def test_create_did_without_endorser_need_to_be(looper, txnPoolNodeSet, nym_txn_data, sdk_pool_handle,
                                                sdk_wallet_trustee):
    change_auth_rule(looper, sdk_pool_handle, sdk_wallet_trustee, constraint=AuthConstraint(role='*',
                                                                                            sig_count=1,
                                                                                            off_ledger_signature=False))

    wh, alias, sender_did, sender_verkey = nym_txn_data
    nym_request = looper.loop.run_until_complete(
        build_nym_request(sender_did, sender_did, sender_verkey, alias, NEW_ROLE))

    request_couple = sdk_sign_and_send_prepared_request(looper, (wh, sender_did), sdk_pool_handle, nym_request)
    with pytest.raises(RequestRejectedException, match='is not found in the Ledger'):
        sdk_get_and_check_replies(looper, [request_couple])



import json
from copy import deepcopy
from hashlib import sha256

from plenum.common.constants import TARGET_NYM, NONCE, RAW, ENC, HASH, NAME, \
    VERSION, FORCE, ORIGIN, OPERATION_SCHEMA_IS_STRICT, OP_VER, TXN_METADATA_SEQ_NO, \
    ALIAS, TXN_TYPE, DATA, VERKEY, ROLE, NYM
from plenum.common.messages.client_request import ClientMessageValidator as PClientMessageValidator
from plenum.common.messages.client_request import ClientOperationField as PClientOperationField
from plenum.common.messages.client_request import ClientNYMOperation as PClientNYMOperation
from plenum.common.messages.fields import ConstantField, IdentifierField, \
    LimitedLengthStringField, TxnSeqNoField, \
    Sha256HexField, JsonField, MapField, BooleanField, VersionField, \
    ChooseField, IntegerField, IterableField, \
    AnyMapField, NonEmptyStringField, DatetimeStringField, RoleField, AnyField, FieldBase, \
    VerkeyField, DestNymField, NonNegativeNumberField
from plenum.common.messages.message_base import MessageValidator
from plenum.common.request import Request as PRequest
from plenum.common.types import OPERATION
from plenum.common.util import is_network_ip_address_valid, is_network_port_valid
from plenum.config import JSON_FIELD_LIMIT, NAME_FIELD_LIMIT, DATA_FIELD_LIMIT, \
    NONCE_FIELD_LIMIT, ALIAS_FIELD_LIMIT, \
    ENC_FIELD_LIMIT, RAW_FIELD_LIMIT, SIGNATURE_TYPE_FIELD_LIMIT
from common.version import GenericVersion
from indy_common.config import DIDDOC_CONTENT_SIZE_LIMIT

from indy_common.authorize.auth_actions import ADD_PREFIX, EDIT_PREFIX
from indy_common.authorize.auth_constraints import ConstraintsEnum, CONSTRAINT_ID, AUTH_CONSTRAINTS, METADATA, \
    NEED_TO_BE_OWNER, SIG_COUNT, ROLE as AUTH_ROLE, OFF_LEDGER_SIGNATURE
from indy_common.config import SCHEMA_ATTRIBUTES_LIMIT
from indy_common.constants import ATTRIB, GET_ATTR, \
    GET_NYM, GET_SCHEMA, GET_CLAIM_DEF, ACTION, \
    POOL_UPGRADE, POOL_CONFIG, \
    DISCLO, SCHEMA, ENDPOINT, CLAIM_DEF, SCHEDULE, NYM_VERSION, SHA256, \
    TIMEOUT, JUSTIFICATION, JUSTIFICATION_MAX_SIZE, REINSTALL, WRITES, START, CANCEL, \
    REVOC_REG_DEF, ISSUANCE_TYPE, MAX_CRED_NUM, PUBLIC_KEYS, \
    TAILS_HASH, TAILS_LOCATION, ID, REVOC_TYPE, TAG, CRED_DEF_ID, VALUE, \
    REVOC_REG_ENTRY, ISSUED, REVOC_REG_DEF_ID, REVOKED, ACCUM, PREV_ACCUM, \
    GET_REVOC_REG_DEF, GET_REVOC_REG, TIMESTAMP, \
    GET_REVOC_REG_DELTA, FROM, TO, POOL_RESTART, DATETIME, VALIDATOR_INFO, \
    SCHEMA_FROM, SCHEMA_NAME, SCHEMA_VERSION, \
    SCHEMA_ATTR_NAMES, CLAIM_DEF_SIGNATURE_TYPE, CLAIM_DEF_PUBLIC_KEYS, CLAIM_DEF_TAG, CLAIM_DEF_SCHEMA_REF, \
    CLAIM_DEF_PRIMARY, CLAIM_DEF_REVOCATION, CLAIM_DEF_FROM, PACKAGE, AUTH_RULE, AUTH_RULES, CONSTRAINT, AUTH_ACTION, \
    AUTH_TYPE, \
    FIELD, OLD_VALUE, NEW_VALUE, GET_AUTH_RULE, RULES, ISSUANCE_BY_DEFAULT, ISSUANCE_ON_DEMAND, RS_TYPE, \
    TAG_LIMIT_SIZE, JSON_LD_CONTEXT, RS_VERSION, \
    RS_NAME, RS_ID, RS_CONTENT, RS_CONTEXT_TYPE_VALUE, RICH_SCHEMA, RS_SCHEMA_TYPE_VALUE, RS_ENCODING_TYPE_VALUE, \
    RICH_SCHEMA_ENCODING, RS_MAPPING_TYPE_VALUE, RICH_SCHEMA_MAPPING, RS_CRED_DEF_TYPE_VALUE, \
    RICH_SCHEMA_CRED_DEF, GET_RICH_SCHEMA_OBJECT_BY_ID, GET_RICH_SCHEMA_OBJECT_BY_METADATA, \
    RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE, DIDDOC_CONTENT
from indy_common.version import SchemaVersion


class Request(PRequest):
    def signingPayloadState(self, identifier=None):
        """
        Special signing state where the the data for an attribute is hashed
        before signing
        :return: state to be used when signing
        """
        if self.operation.get(TXN_TYPE) == ATTRIB:
            d = deepcopy(super().signingPayloadState(identifier=identifier))
            op = d[OPERATION]
            keyName = {RAW, ENC, HASH}.intersection(set(op.keys())).pop()
            op[keyName] = sha256(op[keyName].encode()).hexdigest()
            return d
        return super().signingPayloadState(identifier=identifier)


class ClientGetNymOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(GET_NYM)),
        (TARGET_NYM, IdentifierField()),
        (TIMESTAMP, IntegerField(optional=True)),
        (TXN_METADATA_SEQ_NO, TxnSeqNoField(optional=True)),
    )


class ClientNYMOperation(PClientNYMOperation):
    schema = (
        (TXN_TYPE, ConstantField(NYM)),
        (ALIAS, LimitedLengthStringField(max_length=ALIAS_FIELD_LIMIT, optional=True)),
        (VERKEY, VerkeyField(optional=True, nullable=True)),
        (TARGET_NYM, DestNymField()),
        (ROLE, RoleField(optional=True)),
        (DIDDOC_CONTENT, JsonField(max_length=DIDDOC_CONTENT_SIZE_LIMIT, optional=True)),
        (NYM_VERSION, IntegerField(optional=True)),
    )


class ClientDiscloOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(DISCLO)),
        (DATA, LimitedLengthStringField(max_length=DATA_FIELD_LIMIT)),
        (NONCE, LimitedLengthStringField(max_length=NONCE_FIELD_LIMIT)),
        (TARGET_NYM, IdentifierField(optional=True)),
    )


class GetSchemaField(MessageValidator):
    schema = (
        (SCHEMA_NAME, LimitedLengthStringField(max_length=NAME_FIELD_LIMIT)),
        (SCHEMA_VERSION, VersionField(version_cls=SchemaVersion)),
        (ORIGIN, IdentifierField(optional=True))
    )


class SchemaField(MessageValidator):
    schema = (
        (SCHEMA_NAME, LimitedLengthStringField(max_length=NAME_FIELD_LIMIT)),
        (SCHEMA_VERSION, VersionField(version_cls=SchemaVersion)),
        (SCHEMA_ATTR_NAMES, IterableField(
            LimitedLengthStringField(max_length=NAME_FIELD_LIMIT),
            min_length=1,
            max_length=SCHEMA_ATTRIBUTES_LIMIT)),
    )


class ClaimDefField(MessageValidator):
    schema = (
        (CLAIM_DEF_PRIMARY, AnyMapField()),
        (CLAIM_DEF_REVOCATION, AnyMapField(optional=True)),
    )


class RevocDefValueField(MessageValidator):
    schema = (
        (ISSUANCE_TYPE, ChooseField(values=(ISSUANCE_BY_DEFAULT,
                                            ISSUANCE_ON_DEMAND))),
        (MAX_CRED_NUM, IntegerField()),
        (PUBLIC_KEYS, AnyMapField()),
        (TAILS_HASH, NonEmptyStringField()),
        (TAILS_LOCATION, NonEmptyStringField()),
    )


class ClientRevocDefSubmitField(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(REVOC_REG_DEF)),
        (ID, NonEmptyStringField()),
        (REVOC_TYPE, LimitedLengthStringField()),
        (TAG, LimitedLengthStringField(max_length=TAG_LIMIT_SIZE)),
        (CRED_DEF_ID, NonEmptyStringField()),
        (VALUE, RevocDefValueField())
    )


class RevocRegEntryValueField(MessageValidator):
    schema = (
        (PREV_ACCUM, NonEmptyStringField(optional=True)),
        (ACCUM, NonEmptyStringField()),
        (ISSUED, IterableField(inner_field_type=IntegerField(), optional=True)),
        (REVOKED, IterableField(inner_field_type=IntegerField(), optional=True))
    )


class ClientRevocRegEntrySubmitField(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(REVOC_REG_ENTRY)),
        (REVOC_REG_DEF_ID, NonEmptyStringField()),
        (REVOC_TYPE, LimitedLengthStringField()),
        (VALUE, RevocRegEntryValueField())
    )


class ClientSchemaOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(SCHEMA)),
        (DATA, SchemaField()),
    )


class ClientGetSchemaOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(GET_SCHEMA)),
        (SCHEMA_FROM, IdentifierField()),
        (DATA, GetSchemaField()),
    )


class ClientAttribOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(ATTRIB)),
        (TARGET_NYM, IdentifierField()),
        (RAW, JsonField(max_length=JSON_FIELD_LIMIT, optional=True)),
        (ENC, LimitedLengthStringField(max_length=ENC_FIELD_LIMIT, optional=True)),
        (HASH, Sha256HexField(optional=True)),
    )

    def _validate_message(self, msg):
        self._validate_field_set(msg)
        if RAW in msg:
            self.__validate_raw_field(msg[RAW])

    def _validate_field_set(self, msg):
        fields_n = sum(1 for f in (RAW, ENC, HASH) if f in msg)
        if fields_n == 0:
            self._raise_missed_fields(RAW, ENC, HASH)
        if fields_n > 1:
            self._raise_invalid_message(
                "only one field from {}, {}, {} is expected".format(
                    RAW, ENC, HASH)
            )

    def __validate_raw_field(self, raw_field):
        raw = self.__decode_raw_field(raw_field)
        if not isinstance(raw, dict):
            self._raise_invalid_fields(RAW, type(raw),
                                       'should be a dict')
        if len(raw) != 1:
            self._raise_invalid_fields(RAW, raw,
                                       'should contain one attribute')
        if ENDPOINT in raw:
            self.__validate_endpoint_ha_field(raw[ENDPOINT])

    def __decode_raw_field(self, raw_field):
        return json.loads(raw_field)

    def __validate_endpoint_ha_field(self, endpoint):
        if endpoint is None:
            return  # remove the attribute, valid case
        if not isinstance(endpoint, dict):
            self._raise_invalid_fields(ENDPOINT, endpoint,
                                       'should be a dict')
        HA_NAME = 'ha'
        ha = endpoint.get(HA_NAME)
        if ha is None:
            return  # remove ha attribute, valid case
        if ':' not in ha:
            self._raise_invalid_fields(ENDPOINT, endpoint,
                                       "invalid endpoint format "
                                       "ip_address:port")
        ip, port = ha.split(':')
        if not is_network_ip_address_valid(ip):
            self._raise_invalid_fields('ha', ha,
                                       'invalid endpoint address')
        if not is_network_port_valid(port):
            self._raise_invalid_fields('ha', ha,
                                       'invalid endpoint port')


class ClientGetAttribOperation(ClientAttribOperation):
    schema = (
        (TXN_TYPE, ConstantField(GET_ATTR)),
        (TARGET_NYM, IdentifierField(optional=True)),
        (RAW, LimitedLengthStringField(max_length=RAW_FIELD_LIMIT, optional=True)),
        (ENC, LimitedLengthStringField(max_length=ENC_FIELD_LIMIT, optional=True)),
        (HASH, Sha256HexField(optional=True)),
        (TIMESTAMP, IntegerField(optional=True)),
        (TXN_METADATA_SEQ_NO, TxnSeqNoField(optional=True)),
    )

    def _validate_message(self, msg):
        self._validate_field_set(msg)


class ClientClaimDefSubmitOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(CLAIM_DEF)),
        (CLAIM_DEF_SCHEMA_REF, TxnSeqNoField()),
        (CLAIM_DEF_PUBLIC_KEYS, ClaimDefField()),
        (CLAIM_DEF_SIGNATURE_TYPE, LimitedLengthStringField(max_length=SIGNATURE_TYPE_FIELD_LIMIT)),
        (CLAIM_DEF_TAG, LimitedLengthStringField(max_length=256)),
    )


class ClientClaimDefGetOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(GET_CLAIM_DEF)),
        (CLAIM_DEF_SCHEMA_REF, TxnSeqNoField()),
        (CLAIM_DEF_FROM, IdentifierField()),
        (CLAIM_DEF_SIGNATURE_TYPE, LimitedLengthStringField(max_length=SIGNATURE_TYPE_FIELD_LIMIT)),
        (CLAIM_DEF_TAG, LimitedLengthStringField(max_length=256, optional=True)),
    )


class ClientGetRevocRegDefField(MessageValidator):
    schema = (
        (ID, NonEmptyStringField()),
        (TXN_TYPE, ConstantField(GET_REVOC_REG_DEF)),
    )


class ClientGetRevocRegField(MessageValidator):
    schema = (
        (REVOC_REG_DEF_ID, NonEmptyStringField()),
        (TIMESTAMP, IntegerField()),
        (TXN_TYPE, ConstantField(GET_REVOC_REG)),
    )


class ClientGetRevocRegDeltaField(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(GET_REVOC_REG_DELTA)),
        (REVOC_REG_DEF_ID, NonEmptyStringField()),
        (FROM, IntegerField(optional=True)),
        (TO, IntegerField()),
    )


class ClientPoolUpgradeOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(POOL_UPGRADE)),
        (ACTION, ChooseField(values=(START, CANCEL,))),
        # light check only during static validation
        (VERSION, VersionField(version_cls=GenericVersion)),
        # TODO replace actual checks (idr, datetime)
        (SCHEDULE, MapField(IdentifierField(),
                            NonEmptyStringField(), optional=True)),
        (SHA256, Sha256HexField()),
        (TIMEOUT, NonNegativeNumberField(optional=True)),
        (JUSTIFICATION, LimitedLengthStringField(max_length=JUSTIFICATION_MAX_SIZE, optional=True, nullable=True)),
        (NAME, LimitedLengthStringField(max_length=NAME_FIELD_LIMIT)),
        (FORCE, BooleanField(optional=True)),
        (REINSTALL, BooleanField(optional=True)),
        (PACKAGE, LimitedLengthStringField(max_length=NAME_FIELD_LIMIT, optional=True)),
    )


class ClientPoolRestartOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(POOL_RESTART)),
        (ACTION, ChooseField(values=(START, CANCEL,))),
        (DATETIME, DatetimeStringField(exceptional_values=["0", ""],
                                       optional=True)),
    )


class ClientValidatorInfoOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(VALIDATOR_INFO)),
    )


class ClientPoolConfigOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(POOL_CONFIG)),
        (WRITES, BooleanField()),
        (FORCE, BooleanField(optional=True)),
    )


class ConstraintField(FieldBase):
    _base_types = None

    def __init__(self, constraint_entity, constraint_list, **kwargs):
        super().__init__(**kwargs)
        self._constraint_entity = constraint_entity
        self._constraint_list = constraint_list

    def _specific_validation(self, val):
        if not val:
            return "Fields {} and {} are required and should not " \
                   "be an empty list.".format(AUTH_CONSTRAINTS, CONSTRAINT)
        if CONSTRAINT_ID not in val:
            return "Field {} is required".format(CONSTRAINT_ID)
        if val[CONSTRAINT_ID] == ConstraintsEnum.FORBIDDEN_CONSTRAINT_ID:
            return
        elif val[CONSTRAINT_ID] == ConstraintsEnum.ROLE_CONSTRAINT_ID:
            return self._constraint_entity.validate(val)
        else:
            return self._constraint_list.validate(val)


class ConstraintEntityField(MessageValidator):
    schema = (
        (CONSTRAINT_ID, ChooseField(values=ConstraintsEnum.values())),
        (AUTH_ROLE, RoleField()),
        (SIG_COUNT, NonNegativeNumberField()),
        (NEED_TO_BE_OWNER, BooleanField(optional=True)),
        (OFF_LEDGER_SIGNATURE, BooleanField(optional=True)),
        (METADATA, AnyMapField(optional=True))
    )


class ConstraintListField(MessageValidator):
    schema = ((CONSTRAINT_ID, ChooseField(values=ConstraintsEnum.values())),
              (AUTH_CONSTRAINTS, IterableField(AnyField())))

    def _validate_message(self, val):
        constraints = val.get(AUTH_CONSTRAINTS)
        if not constraints:
            self._raise_invalid_message("Fields {} should not be an empty "
                                        "list.".format(AUTH_CONSTRAINTS))
        for constraint in constraints:
            error_msg = ConstraintField(ConstraintEntityField(),
                                        self).validate(constraint)
            if error_msg:
                self._raise_invalid_message(error_msg)


class AuthRuleValueField(LimitedLengthStringField):
    _base_types = (str, type(None))

    def __init__(self, **kwargs):
        max_length = NAME_FIELD_LIMIT
        super().__init__(max_length, can_be_empty=True, **kwargs)


class AuthRuleField(MessageValidator):
    schema = (
        (CONSTRAINT, ConstraintField(ConstraintEntityField(),
                                     ConstraintListField())),
        (AUTH_ACTION, ChooseField(values=(ADD_PREFIX, EDIT_PREFIX))),
        (AUTH_TYPE, LimitedLengthStringField(max_length=NAME_FIELD_LIMIT)),
        (FIELD, LimitedLengthStringField(max_length=NAME_FIELD_LIMIT)),
        (OLD_VALUE, AuthRuleValueField(optional=True)),
        (NEW_VALUE, AuthRuleValueField())
    )


class ClientAuthRuleOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(AUTH_RULE)),
        *AuthRuleField.schema
    )


class ClientAuthRulesOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(AUTH_RULES)),
        (RULES, IterableField(AuthRuleField(), min_length=1))
    )


class ClientGetAuthRuleOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(GET_AUTH_RULE)),
        (AUTH_ACTION, ChooseField(values=(ADD_PREFIX, EDIT_PREFIX), optional=True)),
        (AUTH_TYPE, LimitedLengthStringField(max_length=NAME_FIELD_LIMIT, optional=True)),
        (FIELD, LimitedLengthStringField(max_length=NAME_FIELD_LIMIT, optional=True)),
        (OLD_VALUE, AuthRuleValueField(optional=True)),
        (NEW_VALUE, AuthRuleValueField(optional=True))
    )


def rich_schema_objects_schema(txn_type, rs_type):
    return (
        (TXN_TYPE, ConstantField(txn_type)),
        (RS_ID, NonEmptyStringField()),
        (RS_TYPE, ConstantField(rs_type)),
        (RS_NAME, LimitedLengthStringField(max_length=NAME_FIELD_LIMIT)),
        (RS_VERSION, VersionField(version_cls=SchemaVersion)),
        (RS_CONTENT, NonEmptyStringField()),
        (OP_VER, LimitedLengthStringField(optional=True))
    )


class ClientJsonLdContextOperation(MessageValidator):
    schema = rich_schema_objects_schema(JSON_LD_CONTEXT, RS_CONTEXT_TYPE_VALUE)


class ClientRichSchemaOperation(MessageValidator):
    schema = rich_schema_objects_schema(RICH_SCHEMA, RS_SCHEMA_TYPE_VALUE)


class ClientRichSchemaEncodingOperation(MessageValidator):
    schema = rich_schema_objects_schema(RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE)


class ClientRichSchemaMappingOperation(MessageValidator):
    schema = rich_schema_objects_schema(RICH_SCHEMA_MAPPING, RS_MAPPING_TYPE_VALUE)


class ClientRichSchemaCredDefOperation(MessageValidator):
    schema = rich_schema_objects_schema(RICH_SCHEMA_CRED_DEF, RS_CRED_DEF_TYPE_VALUE)


class ClientRichSchemaPresDefOperation(MessageValidator):
    schema = rich_schema_objects_schema(RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE)


class ClientGetRichSchemaObjectByIdOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(GET_RICH_SCHEMA_OBJECT_BY_ID)),
        (RS_ID, NonEmptyStringField()),
    )


class ClientGetRichSchemaObjectByMetadataOperation(MessageValidator):
    schema = (
        (TXN_TYPE, ConstantField(GET_RICH_SCHEMA_OBJECT_BY_METADATA)),
        (RS_TYPE, NonEmptyStringField()),
        (RS_NAME, LimitedLengthStringField(max_length=NAME_FIELD_LIMIT)),
        (RS_VERSION, VersionField(version_cls=SchemaVersion)),
    )


class ClientOperationField(PClientOperationField):
    _specific_operations = {
        NYM: ClientNYMOperation(),
        SCHEMA: ClientSchemaOperation(),
        ATTRIB: ClientAttribOperation(),
        GET_ATTR: ClientGetAttribOperation(),
        CLAIM_DEF: ClientClaimDefSubmitOperation(),
        GET_CLAIM_DEF: ClientClaimDefGetOperation(),
        DISCLO: ClientDiscloOperation(),
        GET_NYM: ClientGetNymOperation(),
        GET_SCHEMA: ClientGetSchemaOperation(),
        POOL_UPGRADE: ClientPoolUpgradeOperation(),
        POOL_CONFIG: ClientPoolConfigOperation(),
        AUTH_RULE: ClientAuthRuleOperation(),
        AUTH_RULES: ClientAuthRulesOperation(),
        GET_AUTH_RULE: ClientGetAuthRuleOperation(),
        POOL_RESTART: ClientPoolRestartOperation(),
        VALIDATOR_INFO: ClientValidatorInfoOperation(),
        REVOC_REG_DEF: ClientRevocDefSubmitField(),
        REVOC_REG_ENTRY: ClientRevocRegEntrySubmitField(),
        GET_REVOC_REG_DEF: ClientGetRevocRegDefField(),
        GET_REVOC_REG: ClientGetRevocRegField(),
        GET_REVOC_REG_DELTA: ClientGetRevocRegDeltaField(),
        JSON_LD_CONTEXT: ClientJsonLdContextOperation(),
        RICH_SCHEMA: ClientRichSchemaOperation(),
        RICH_SCHEMA_ENCODING: ClientRichSchemaEncodingOperation(),
        RICH_SCHEMA_MAPPING: ClientRichSchemaMappingOperation(),
        RICH_SCHEMA_CRED_DEF: ClientRichSchemaCredDefOperation(),
        RICH_SCHEMA_PRES_DEF: ClientRichSchemaPresDefOperation(),
        GET_RICH_SCHEMA_OBJECT_BY_ID: ClientGetRichSchemaObjectByIdOperation(),
        GET_RICH_SCHEMA_OBJECT_BY_METADATA: ClientGetRichSchemaObjectByMetadataOperation()
    }

    # TODO: it is a workaround because INDY-338, `operations` must be a class
    # constant
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.operations.update(self._specific_operations)


class ClientMessageValidator(PClientMessageValidator):
    # extend operation field
    schema = tuple(
        map(lambda x: (x[0], ClientOperationField()) if x[0] == OPERATION else x,
            PClientMessageValidator.schema)
    )


# THE CODE BELOW MIGHT BE NEEDED IN THE FUTURE, THEREFORE KEEPING IT
# class LedgerIdField(PLedgerIdField):
#     ledger_ids = PLedgerIdField.ledger_ids + (CONFIG_LEDGER_ID,)
#
#
# class LedgerInfoField(PLedgerInfoField):
#     _ledger_id_class = LedgerIdField


# TODO: it is a workaround which helps extend some fields from
# downstream projects, should be removed after we find a better way
# to do this
# node_message_factory.update_schemas_by_field_type(
#     PLedgerIdField, LedgerIdField)
# node_message_factory.update_schemas_by_field_type(
#     PLedgerInfoField, LedgerInfoField)


class SafeRequest(Request, ClientMessageValidator):
    def __init__(self, **kwargs):
        ClientMessageValidator.__init__(self,
                                        operation_schema_is_strict=OPERATION_SCHEMA_IS_STRICT)
        self.validate(kwargs)
        Request.__init__(self, **kwargs)



import pytest

from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER


@pytest.fixture(scope='module')
def write_auth_req_validator(write_auth_req_validator, key):
    write_auth_req_validator.auth_cons_strategy.get_auth_constraint = lambda a: AuthConstraint(IDENTITY_OWNER, 5)
    return write_auth_req_validator


def test_claim_def_adding_success_5_owners(write_request_validation, req,
                                           identity_owners, endorsers, key, write_auth_req_validator):
    req.signatures = {idr: "signature" for idr in identity_owners}
    # Endorser must be present in case of multi-sig
    req.endorser = endorsers[0]
    req.signatures.update({req.endorser: "endorser_sig"})
    assert write_request_validation(req, [key])


def test_claim_def_adding_fail_4_owners(write_request_validation, req,
                                        identity_owners, key):
    req.signatures = {idr: "signature" for idr in identity_owners[:4]}
    assert not write_request_validation(req, [key])


def test_claim_def_adding_fail_1_owner_4_unknown(write_request_validation, req,
                                                 identity_owners, key):
    req.signatures = {identity_owners[0]: "signature"}
    assert not write_request_validation(req, [key])

    owners_count = len(identity_owners)
    req.signatures.update({"unknown_idr_{}".format(i): "signature"
                           for i in range(owners_count, owners_count + 4)})
    assert not write_request_validation(req, [key])


def test_claim_def_adding_fail_5_trustees(write_request_validation, req,
                                          trustees, key):
    req.signatures = {idr: "signature" for idr in trustees}
    assert not write_request_validation(req, [key])



import json
import pytest

from indy.ledger import build_pool_config_request

from indy_node.test.nym_txn.test_nym_additional import get_nym
from indy_node.test.upgrade.helper import sdk_ensure_upgrade_sent
from indy_node.test.pool_config.helper import sdk_ensure_pool_config_sent

from plenum.common.exceptions import RequestNackedException
from plenum.common.types import OPERATION
from plenum.test.helper import sdk_get_bad_response, sdk_sign_and_submit_req
from plenum.common.constants import VERSION
from plenum.test.pool_transactions.helper import sdk_add_new_nym

from indy_node.test.upgrade.conftest import validUpgrade, nodeIds, pckg


def sdk_pool_bad_config_sent(looper, sdk_pool_handle, sdk_wallet_trustee, change_writes,
                             change_force, change_writes_value=None, change_force_value=None):
    _, did = sdk_wallet_trustee
    req = looper.loop.run_until_complete(build_pool_config_request(
        did, True, True))
    req = json.loads(req)
    del req[OPERATION]['writes']
    req[OPERATION][change_writes] = change_writes_value if change_writes_value else True
    del req[OPERATION]['force']
    req[OPERATION][change_force] = change_force_value if change_force_value else True
    req = json.dumps(req)
    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, req)
    return req


def testPoolConfigInvalidSyntax(looper, sdk_pool_handle, sdk_wallet_trustee, poolConfigWTFF):
    req = sdk_pool_bad_config_sent(looper, sdk_pool_handle, sdk_wallet_trustee,
                                   'wites', 'force', True, False)
    sdk_get_bad_response(looper, [req], RequestNackedException, 'missed fields - writes')
    req = sdk_pool_bad_config_sent(looper, sdk_pool_handle, sdk_wallet_trustee,
                                   'writes', 'force', 'Tue', False)
    sdk_get_bad_response(looper, [req], RequestNackedException, 'expected types \'bool\', got \'str\'')
    req = sdk_pool_bad_config_sent(looper, sdk_pool_handle, sdk_wallet_trustee,
                                   'writes', 'force', True, 1)
    sdk_get_bad_response(looper, [req], RequestNackedException, 'expected types \'bool\', got \'int\'')


def testPoolConfigWritableFalse(looper, sdk_pool_handle, sdk_wallet_trustee, poolConfigWFFF):
    sdk_ensure_pool_config_sent(looper, sdk_pool_handle, sdk_wallet_trustee,
                                poolConfigWFFF)
    with pytest.raises(RequestNackedException) as e:
        sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    e.match('Pool is in readonly mode')


def testPoolConfigWritableTrue(looper, sdk_pool_handle, sdk_wallet_trustee, poolConfigWTFF):
    with pytest.raises(RequestNackedException) as e:
        sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    e.match('Pool is in readonly mode')
    sdk_ensure_pool_config_sent(looper, sdk_pool_handle, sdk_wallet_trustee,
                                poolConfigWTFF)
    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)


def testPoolConfigWritableFalseCanRead(looper, sdk_pool_handle, sdk_wallet_trustee, poolConfigWFFF):
    _, did = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    get_nym(looper, sdk_pool_handle, sdk_wallet_trustee, did)
    sdk_ensure_pool_config_sent(looper, sdk_pool_handle, sdk_wallet_trustee,
                                poolConfigWFFF)
    with pytest.raises(RequestNackedException) as e:
        sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    e.match('Pool is in readonly mode')
    get_nym(looper, sdk_pool_handle, sdk_wallet_trustee, did)


def testPoolUpgradeOnReadonlyPool(
        looper, nodeSet, sdk_pool_handle, sdk_wallet_trustee, validUpgrade, poolConfigWFFF):
    sdk_ensure_pool_config_sent(looper, sdk_pool_handle, sdk_wallet_trustee,
                                poolConfigWFFF)
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee,
                     validUpgrade)

    for node in nodeSet:
        assert len(node.upgrader.aqStash) > 0
        assert node.upgrader.scheduledAction
        assert node.upgrader.scheduledAction.version.full == validUpgrade[VERSION]



from operator import itemgetter
from indy_common.util import getIndex


def test_getIndex():
    items = [('a', {'key1': 1}), ('b', {'key2': 2})]
    getDict = itemgetter(1)

    def containsKey(key):
        return lambda x: key in getDict(x)

    assert 0 == getIndex(containsKey('key1'), items)
    assert 1 == getIndex(containsKey('key2'), items)
    assert -1 == getIndex(containsKey('key3'), items)



from common.serializers.json_serializer import JsonSerializer
from indy_common.authorize.auth_request_validator import WriteRequestValidator

from indy_common.constants import RICH_SCHEMA_CRED_DEF, RS_CRED_DEF_SIG_TYPE, RS_CRED_DEF_MAPPING, \
    RS_CRED_DEF_SCHEMA, RS_CRED_DEF_PUB_KEY, RS_CONTENT, RS_TYPE, RS_SCHEMA_TYPE_VALUE, RS_MAPPING_TYPE_VALUE

from indy_node.server.request_handlers.domain_req_handlers.rich_schema.abstract_rich_schema_object_handler import \
    AbstractRichSchemaObjectHandler
from plenum.common.exceptions import InvalidClientRequest

from plenum.server.database_manager import DatabaseManager
from stp_core.common.log import getlogger

logger = getlogger()


class RichSchemaCredDefHandler(AbstractRichSchemaObjectHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(RICH_SCHEMA_CRED_DEF, database_manager, write_req_validator)

    def is_json_ld_content(self):
        return False

    def do_static_validation_content(self, content_as_dict, request):
        missing_fields = []
        for field in [RS_CRED_DEF_SIG_TYPE, RS_CRED_DEF_MAPPING, RS_CRED_DEF_SCHEMA, RS_CRED_DEF_PUB_KEY]:
            if not content_as_dict.get(field):
                missing_fields.append("'{}'".format(field))

        if missing_fields:
            missing_fields_str = ", ".join(missing_fields)
            raise InvalidClientRequest(request.identifier, request.reqId,
                                       "{} must be set in '{}'".format(missing_fields_str, RS_CONTENT))

    def do_dynamic_validation_content(self, request):
        # it has been checked on static validation step that the content is a valid JSON.
        # and it has schema and mapping fields
        content_as_dict = JsonSerializer.loads(request.operation[RS_CONTENT])
        schema_id = content_as_dict[RS_CRED_DEF_SCHEMA]
        mapping_id = content_as_dict[RS_CRED_DEF_MAPPING]

        # 1. check that the schema field points to an existing object on the ledger
        schema, _, _ = self.get_from_state(schema_id)
        if not schema:
            raise InvalidClientRequest(request.identifier,
                                       request.reqId,
                                       "Can not find a referenced '{}' with id={}; please make sure that it has been added to the ledger".format(
                                           RS_CRED_DEF_SCHEMA, schema_id))

        # 2. check that the mapping field points to an existing object on the ledger
        mapping, _, _ = self.get_from_state(mapping_id)
        if not mapping:
            raise InvalidClientRequest(request.identifier,
                                       request.reqId,
                                       "Can not find a referenced '{}' with id={}; please make sure that it has been added to the ledger".format(
                                           RS_CRED_DEF_MAPPING, mapping_id))

        # 3. check that the schema field points to an object of the Schema type
        if schema.get(RS_TYPE) != RS_SCHEMA_TYPE_VALUE:
            raise InvalidClientRequest(request.identifier,
                                       request.reqId,
                                       "'{}' field must reference a schema with {}={}".format(
                                           RS_CRED_DEF_SCHEMA, RS_TYPE, RS_SCHEMA_TYPE_VALUE))

        # 4. check that the mapping fields points to an object of the Mapping type
        if mapping.get(RS_TYPE) != RS_MAPPING_TYPE_VALUE:
            raise InvalidClientRequest(request.identifier,
                                       request.reqId,
                                       "'{}' field must reference a mapping with {}={}".format(
                                           RS_CRED_DEF_MAPPING, RS_TYPE, RS_MAPPING_TYPE_VALUE))



import indy_node.utils.migration_tool as migration_tool
import pytest
from timeout_decorator import TimeoutError


TEST_MIGRATION_SCRIPTS = [
    '0_3_100_to_0_3_104',
    '10_2_1_to_10_0_2',
    '0_9_2_to_0_9_10',
    '0_3_100_to_0_3_101',
    '0_1_10_to_0_9_10',
    '0_3_4_invalid']
TEST_VERSION = '0.3.100'
TEST_NEW_VERSION = '0.10.105'
TEST_TIMEOUT = 1


def testGetRelevantMigrations():
    relevantMigrations = migration_tool._get_relevant_migrations(
        TEST_MIGRATION_SCRIPTS, TEST_VERSION, TEST_NEW_VERSION)
    assert len(relevantMigrations) == 3
    assert relevantMigrations == [
        '0_3_100_to_0_3_101', '0_3_100_to_0_3_104', '0_9_2_to_0_9_10']


def testMigrate(monkeypatch):
    testList = []

    monkeypatch.setattr(migration_tool, '_call_migration_script',
                        lambda *x: testList.append(x[0]))
    monkeypatch.setattr(migration_tool, '_get_migration_scripts',
                        lambda *x: TEST_MIGRATION_SCRIPTS)

    assert migration_tool.migrate(
        TEST_VERSION, TEST_NEW_VERSION, TEST_TIMEOUT) == 3
    assert len(testList) == 3


def testMigrateTimesOut(monkeypatch):
    monkeypatch.setattr(migration_tool, '_call_migration_script',
                        lambda *x: exec('raise(TimeoutError())'))
    monkeypatch.setattr(migration_tool, '_get_migration_scripts',
                        lambda *x: TEST_MIGRATION_SCRIPTS)

    with pytest.raises(TimeoutError):
        migration_tool.migrate(TEST_VERSION, TEST_NEW_VERSION, TEST_TIMEOUT)

def testGetMigrationScripts():
    try:
        res = migration_tool._get_migration_scripts(migration_tool._get_current_platform())
    except Exception as e:
        pytest.fail("Unexpected error: {}".format(e))
    else:
        assert type(res) is list
        assert res



import multiprocessing
from stp_core.loop.eventually import eventually
from indy_common.version import src_version_cls
from indy_node.utils.node_control_utils import NodeControlUtil, DebianVersion

from indy_node.test.upgrade.helper import (
    NodeControlToolExecutor as NCT,
    composeUpgradeMessage,
    sendUpgradeMessage,
    nodeControlGeneralMonkeypatching,
    bumpedVersion
)


m = multiprocessing.Manager()
# TODO why do we expect that
whitelist = ['Unexpected error in _upgrade test']


def testNodeControlReceivesMessages(monkeypatch, looper, tdir, tconf):
    received = m.list()
    version = bumpedVersion()
    stdout = 'teststdout'

    def transform(tool):
        nodeControlGeneralMonkeypatching(tool, monkeypatch, tdir, stdout)
        monkeypatch.setattr(tool, '_process_data', received.append)

    def checkMessage():
        assert len(received) == 1
        assert received[0] == composeUpgradeMessage(version)

    monkeypatch.setattr(
        NodeControlUtil, 'get_latest_pkg_version',
        lambda *x, **y: DebianVersion(
            version, upstream_cls=src_version_cls())
    )

    nct = NCT(backup_dir=tdir, backup_target=tdir, transform=transform)
    try:
        sendUpgradeMessage(version)
        looper.run(eventually(checkMessage))
    finally:
        nct.stop()



import logging
from collections import OrderedDict

from indy_common.constants import CONFIG_LEDGER_AUTH_POLICY
from plenum.common.constants import ClientBootStrategy, HS_ROCKSDB, KeyValueStorageType
from plenum.config import rocksdb_default_config

nodeReg = OrderedDict([
    ('Alpha', ('127.0.0.1', 9701)),
    ('Beta', ('127.0.0.1', 9703)),
    ('Gamma', ('127.0.0.1', 9705)),
    ('Delta', ('127.0.0.1', 9707))
])

cliNodeReg = OrderedDict([
    ('AlphaC', ('127.0.0.1', 9702)),
    ('BetaC', ('127.0.0.1', 9704)),
    ('GammaC', ('127.0.0.1', 9706)),
    ('DeltaC', ('127.0.0.1', 9708))
])

GENERAL_CONFIG_DIR = "/etc/indy"

GENERAL_CONFIG_FILE = 'indy_config.py'
NETWORK_CONFIG_FILE = 'indy_config.py'
USER_CONFIG_FILE = 'indy_config.py'

configTransactionsFile = "config_transactions"

logFilePath = "cli.log"

outFilePath = "cli_output.log"

clientBootStrategy = ClientBootStrategy.Custom

hashStore = {
    "type": HS_ROCKSDB
}

primaryStorage = None

configStateStorage = KeyValueStorageType.Rocksdb
idrCacheStorage = KeyValueStorageType.Rocksdb
attrStorage = KeyValueStorageType.Rocksdb

configStateDbName = 'config_state'
attrDbName = 'attr_db'
idrCacheDbName = 'idr_cache_db'

rocksdb_attr_db_config = rocksdb_default_config.copy()
# Change attr_db config here if you fully understand what's going on

rocksdb_idr_cache_db_config = rocksdb_default_config.copy()
# Change idr_cache_db config here if you fully understand what's going on

db_attr_db_config = rocksdb_attr_db_config
db_idr_cache_db_config = rocksdb_idr_cache_db_config

PluginsToLoad = []

# TODO: This should be in indy_node's config

# File that stores the version of the Node ran the last time it started. (It
# might be incorrect sometimes if Node failed to update the file and crashed)
lastRunVersionFile = 'last_version'

# File that stores the version of the code to which the update has to be made.
# This is used to detect if there was an error while upgrading. Once it has
# been found out that there was error while upgrading, then it can be upgraded.
nextVersionFile = 'next_version'

# Minimum time difference (seconds) between the code update of 2 nodes
MinSepBetweenNodeUpgrades = 300

upgradeLogFile = "upgrade_log"
restartLogFile = "restart_log"

lastVersionFilePath = "last_version_file"

'''
Node control utils options
'''
controlServiceHost = "127.0.0.1"
controlServicePort = "30003"

'''
logging level for agents
'''
agentLoggingLevel = logging.INFO
'''
default logging level for node
'''
logLevel = logging.INFO

INCONSISTENCY_WATCHER_NETWORK_TIMEOUT = 90

# Top level package to be updated via pool upgrade command
UPGRADE_ENTRY = 'indy-node'

PACKAGES_TO_HOLD = ['indy-plenum', 'indy-node',
                    # From indy-plenum:
                    'python3-ioflo', 'python3-orderedset', 'python3-base58', 'python3-prompt-toolkit', 'python3-rlp', 'python3-sha3',
                    'python3-libnacl', 'python3-six', 'python3-portalocker', 'python3-sortedcontainers',
                    'python3-dateutil', 'python3-semver', 'python3-psutil', 'python3-pyzmq',
                    'python3-jsonpickle', 'python3-rocksdb', 'python3-pympler', 'python3-packaging',
                    # From indy-node:
                    'python3-timeout-decorator', 'python3-distro']


authPolicy = CONFIG_LEDGER_AUTH_POLICY

SCHEMA_ATTRIBUTES_LIMIT = 125

CONTEXT_SIZE_LIMIT = 131072
JSON_LD_LIMIT = CONTEXT_SIZE_LIMIT
DIDDOC_CONTENT_SIZE_LIMIT = 10 * 1024

ENABLE_RICH_SCHEMAS = False
